from __future__ import annotations
import asyncio
import traceback
from datetime import datetime
from neo3.network import encode_base62, message, payloads, capabilities, protocol as protocol, relaycache
from neo3.network.ipfilter import ipfilter
from neo3.network.convenience import nodeweight
from neo3 import network_logger as logger, settings, blockchain
from neo3.core import types, msgrouter
from contextlib import suppress
from socket import AF_INET as IP4_FAMILY
from typing import Optional, List, Tuple, Dict, Callable, cast


class NeoNode:

    #: List[payloads.NetworkAddress]: a list of known network addresses (class attribute).
    addresses = []  # type: List[payloads.NetworkAddress]

    def __init__(self, protocol):
        self.protocol = protocol
        #: payloads.NetworkAddress: Address of the remote endpoint.
        self.address = payloads.NetworkAddress("0.0.0.0:0", state=payloads.AddressState.DEAD)
        self.nodeid: int = id(self)  #: int: Unique identifier.
        self.nodeid_human: str = encode_base62(self.nodeid)  #: str: Human readable id.
        self.version = None
        self.tasks = []
        self.nodeweight = nodeweight.NodeWeight(self.nodeid)
        self.best_height: int = 0  #: int: Best block height of node.
        self.best_height_last_update = datetime.utcnow().timestamp()

        self._read_task = None  # type: asyncio.Task
        #: bool: Whether the node is in the process of disconnecting and shutting down its tasks.
        self.disconnecting: bool = False

        #: Dict[message.MessageType, Callable[[message.Message], None]]: A table matching message types to handler
        #: functions.
        self.dispatch_table: Dict[message.MessageType, Callable[[message.Message], None]] = {
            message.MessageType.ADDR: self.handler_addr,
            message.MessageType.BLOCK: self.handler_block,
            message.MessageType.CONSENSUS: self.handler_consensus,
            message.MessageType.INV: self.handler_inv,
            message.MessageType.FILTERADD: self.handler_filteradd,
            message.MessageType.FILTERCLEAR: self.handler_filterclear,
            message.MessageType.FILTERLOAD: self.handler_filterload,
            message.MessageType.GETADDR: self.handler_getaddr,
            message.MessageType.GETBLOCKS: self.handler_getblocks,
            message.MessageType.GETBLOCKBYINDEX: self.handler_getblockdata,
            message.MessageType.GETDATA: self.handler_getdata,
            message.MessageType.GETHEADERS: self.handler_getheaders,
            message.MessageType.HEADERS: self.handler_headers,
            message.MessageType.MEMPOOL: self.handler_mempool,
            message.MessageType.MERKLEBLOCK: self.handler_merkleblock,
            message.MessageType.PING: self.handler_ping,
            message.MessageType.PONG: self.handler_pong,
            message.MessageType.TRANSACTION: self.handler_transaction,
            message.MessageType.EXTENSIBLE: self.handler_extensible
        }

    # connection setup and control functions
    async def connection_made(self, transport) -> None:
        """
        Event called by the :meth:`base protocol <asyncio.BaseProtocol.connection_made>`.
        """
        addr_tuple = self.protocol._stream_writer.get_extra_info('peername')
        address = f"{addr_tuple[0]}:{addr_tuple[1]}"

        network_addr = self._find_address_by_host_port(address)
        if network_addr:
            # this scenario occurs when the NodeManager queues seed nodes
            self.address = network_addr
        else:
            self.address.address = address

        if not ipfilter.is_allowed(addr_tuple[0]):
            logger.debug(f"Blocked by ipfilter: {self.address.address}")
            await self.disconnect(payloads.DisconnectReason.IPFILTER_NOT_ALLOWED)
            return

    async def _do_handshake(self) -> Tuple[bool, Optional[payloads.DisconnectReason]]:
        caps: List[capabilities.NodeCapability] = [capabilities.FullNodeCapability(0)]
        # TODO: fix nonce and port if a service is running
        send_version = message.Message(msg_type=message.MessageType.VERSION,
                                       payload=payloads.VersionPayload(nonce=123,
                                                                       user_agent="NEO-MAMBA",
                                                                       capabilities=caps))
        await self.send_message(send_version)

        m = await self.read_message(timeout=3)
        if not m or m.type != message.MessageType.VERSION:
            await self.disconnect(payloads.DisconnectReason.HANDSHAKE_VERSION_ERROR)
            return (False, payloads.DisconnectReason.HANDSHAKE_VERSION_ERROR)

        if not self._validate_version(m.payload):
            await self.disconnect(payloads.DisconnectReason.HANDSHAKE_VERSION_ERROR)
            return (False, payloads.DisconnectReason.HANDSHAKE_VERSION_ERROR)

        m_verack = message.Message(msg_type=message.MessageType.VERACK)
        await self.send_message(m_verack)

        m = await self.read_message(timeout=3)
        if not m or m.type != message.MessageType.VERACK:
            await self.disconnect(payloads.DisconnectReason.HANDSHAKE_VERACK_ERROR)
            return (False, payloads.DisconnectReason.HANDSHAKE_VERACK_ERROR)

        logger.debug(f"Connected to {self.version.user_agent} @ {self.address.address}: {self.best_height}.")
        msgrouter.on_node_connected(self)

        return (True, None)

    async def disconnect(self, reason: payloads.DisconnectReason) -> None:
        """
        Close the connection to remote endpoint.

        Args:
            reason: reason for disconnecting.
        """
        if self.disconnecting:
            return

        self.disconnecting = True

        logger.debug(f"Disconnect called with reason={reason.name}")
        self.address.disconnect_reason = reason
        if reason in [payloads.DisconnectReason.MAX_CONNECTIONS_REACHED,
                      payloads.DisconnectReason.POOR_PERFORMANCE,
                      payloads.DisconnectReason.HANDSHAKE_VERACK_ERROR,
                      payloads.DisconnectReason.HANDSHAKE_VERSION_ERROR,
                      payloads.DisconnectReason.UNKNOWN]:
            self.address.set_state_poor()
        elif reason == payloads.DisconnectReason.IPFILTER_NOT_ALLOWED:
            self.address.set_state_dead()

        for t in self.tasks:
            t.cancel()
            with suppress(asyncio.CancelledError):
                print(f"waiting for task to cancel {t}.")
                await t
                print("done")
        msgrouter.on_node_disconnected(self, reason)
        self.protocol.disconnect()

    async def connection_lost(self, exc) -> None:
        """
        Event called by the :meth:`base protocol <asyncio.BaseProtocol.connection_lost>`.
        """
        logger.debug(f"{datetime.now()} Connection lost {self.address} exception: {exc}")

        if self.address.is_state_connected:
            await self.disconnect(payloads.DisconnectReason.UNKNOWN)

    def _validate_version(self, version) -> bool:
        if version.nonce == self.nodeid:
            logger.debug("Client is self.")
            return False

        if version.magic != settings.network.magic:
            logger.debug(f"Wrong network id {version.magic}.")
            return False

        for c in version.capabilities:
            if isinstance(c, capabilities.ServerCapability):
                addr = self._find_address_by_host_port(self.address.address)

                if addr:
                    addr.set_state_connected()
                    addr.capabilities = version.capabilities
                else:
                    logger.debug(f"Adding address from outside {self.address.address}.")
                    # new connection initiated from outside
                    addr = payloads.address.NetworkAddress(
                        address=self.address.address,
                        capabilities=version.capabilities,
                        state=payloads.address.AddressState.CONNECTED
                    )
                    self.addresses.append(addr)
                break

        for c in version.capabilities:
            if isinstance(c, capabilities.FullNodeCapability):
                # update nodes height indicator
                self.best_height = c.start_height
                self.best_height_last_update = datetime.utcnow().timestamp()
                self.version = version
                return True
        else:
            return False

    def handler_addr(self, msg: message.Message) -> None:
        """
        Handler for a message with the ADDR type.

        Args:
            msg:
        """
        payload = cast(payloads.AddrPayload, msg.payload)
        self.addresses = list(set(self.addresses + payload.addresses))
        msgrouter.on_addr(payload.addresses)

    def handler_block(self, msg: message.Message) -> None:
        """
        Handler for a message with the BLOCK type.

        Args:
            msg:
        """
        msgrouter.on_block(self.nodeid, msg.payload)

    def handler_consensus(self, msg: message.Message) -> None:
        """
        Handler for a message with the CONSENSUS type.

        Args:
            msg:
        """
        pass

    def handler_inv(self, msg: message.Message) -> None:
        """
        Handler for a message with the INV type.

        Args:
            msg:
        """
        payload = cast(payloads.InventoryPayload, msg.payload)
        if payload.type == payloads.InventoryType.BLOCK:
            # neo-cli broadcasts INV messages on a regular interval. We can use those as trigger to request
            # their latest block height
            if len(payload.hashes) > 0:
                if settings.database:
                    height = max(0, blockchain.Blockchain().height)
                else:
                    height = 0
                m = message.Message(msg_type=message.MessageType.PING,
                                    payload=payloads.PingPayload(height=height))
                self._create_task_with_cleanup(self.send_message(m))
        else:
            logger.debug(f"Message with type INV received. No processing for payload type "  # type:ignore
                         f"{payload.type.name} implemented")

    def handler_filteradd(self, msg: message.Message) -> None:
        """
        Handler for a message with the FILTERADD type.

        Args:
            msg:
        """
        pass

    def handler_filterclear(self, msg: message.Message) -> None:
        """
        Handler for a message with the FILTERCLEAR type.

        Args:
            msg:
        """
        pass

    def handler_filterload(self, msg: message.Message) -> None:
        """
        Handler for a message with the FILTERLOAD type.

        Args:
            msg:
        """
        pass

    def handler_getaddr(self, msg: message.Message) -> None:
        """
        Handler for a message with the GETADDR type.

        Args:
            msg:
        """
        addr_list = []
        for address in self.addresses:  # type: payloads.NetworkAddress
            if address.is_state_new or address.is_state_connected:
                addr_list.append(address)
        self._create_task_with_cleanup(self.send_address_list(addr_list))

    def handler_getblocks(self, msg: message.Message) -> None:
        """
        Handler for a message with the GETBLOCKS type.

        Args:
            msg:
        """
        pass

    def handler_getblockdata(self, msg: message.Message) -> None:
        """
        Handler for a message with the GETBLOCKBYINDEX type.

        Args:
            msg:
        """
        pass

    def handler_getdata(self, msg: message.Message) -> None:
        """
        Handler for a message with the GETDATA type.

        Args:
            msg:
        """
        payload = cast(payloads.InventoryPayload, msg.payload)
        for h in payload.hashes:
            item = relaycache.RelayCache().try_get(h)
            if item is None:
                # for the time being we only support data retrieval for our own relays
                continue
            if payload.type == payloads.InventoryType.TX:
                m = message.Message(msg_type=message.MessageType.TRANSACTION,
                                    payload=item)
                self._create_task_with_cleanup(self.send_message(m))

    def handler_getheaders(self, msg: message.Message) -> None:
        """
        Handler for a message with the GETHEADERS type.

        Args:
            msg:
        """
        pass

    def handler_mempool(self, msg: message.Message) -> None:
        """
        Handler for a message with the MEMPOOL type.

        Args:
            msg:
        """
        pass

    def handler_merkleblock(self, msg: message.Message) -> None:
        """
        Handler for a message with the MERKLEBLOCK type.

        Args:
            msg:
        """
        pass

    def handler_headers(self, msg: message.Message) -> None:
        """
        Handler for a message with the HEADERS type.

        Args:
            msg:
        """
        payload = cast(payloads.HeadersPayload, msg.payload)
        if len(payload.headers) > 0:
            msgrouter.on_headers(self.nodeid, payload.headers)

    def handler_ping(self, msg: message.Message) -> None:
        """
        Handler for a message with the PING type.

        Args:
            msg:
        """
        if settings.database:
            height = max(0, blockchain.Blockchain().height)
        else:
            height = 0
        m = message.Message(msg_type=message.MessageType.PONG,
                            payload=payloads.PingPayload(height=height))
        self._create_task_with_cleanup(self.send_message(m))

    def handler_pong(self, msg: message.Message) -> None:
        """
        Handler for a message with the PONG type.

        Args:
            msg:
        """
        payload = cast(payloads.PingPayload, msg.payload)
        if self.best_height != payload.current_height:
            logger.debug(f"Updating node {self.nodeid_human} height "
                         f"from {self.best_height} to {payload.current_height}")
            self.best_height = payload.current_height
            self.best_height_last_update = datetime.utcnow().timestamp()

    def handler_transaction(self, msg: message.Message) -> None:
        """
        Handler for a message with the TRANSACTION type.

        Args:
            msg:
        """
        pass

    def handler_extensible(self, msg: message.Message) -> None:
        """
        Handler for a message with the EXTENSIBLE type.

        Args:
            msg:
        """
        pass

    async def _process_incoming_data(self) -> None:
        """
        Main loop
        """
        logger.debug("Waiting for a message.")
        while not self.disconnecting:
            # we want to always listen for an incoming message
            m = await self.read_message(timeout=1)
            if m is None:
                continue

            handler = self.dispatch_table.get(m.type, None)
            if handler:
                handler(m)
            else:
                logger.debug(f"Unknown message with type: {m.type.name}.")

    # raw network commands
    async def request_address_list(self) -> None:
        """
        Send a request for receiving known network addresses.
        """
        m = message.Message(msg_type=message.MessageType.GETADDR)
        await self.send_message(m)

    async def send_address_list(self, network_addresses: List[payloads.NetworkAddress]) -> None:
        """
        Send network addresses.

        Args:
            network_addresses:
        """
        m = message.Message(msg_type=message.MessageType.ADDR,
                            payload=payloads.AddrPayload(addresses=network_addresses))
        await self.send_message(m)

    async def request_headers(self, index_start: int, count: int = payloads.HeadersPayload.MAX_HEADERS_COUNT) -> None:
        """
        Send a request for headers from `index_start` to `index_start`+`count`.

        Not specifying a `count` results in requesting at most 2000 headers.

        Args:
            index_start:
            count:
        """
        m = message.Message(msg_type=message.MessageType.GETHEADERS,
                            payload=payloads.GetBlockByIndexPayload(index_start, count))
        await self.send_message(m)

    async def send_headers(self, headers: List[payloads.Header]) -> None:
        """
        Send a list of Header objects.

        Args:
            headers:
        """
        if len(headers) > 2000:
            headers = headers[:2000]

        m = message.Message(msg_type=message.MessageType.HEADERS, payload=payloads.HeadersPayload(headers))
        await self.send_message(m)

    async def request_blocks(self, hash_start: types.UInt256, count: int = None) -> None:
        """
        Send a request for retrieving block hashes from `hash_start` to `hash_start`+`count`.

        Not specifying a `count` results in requesting at most 500 blocks.

        Note:
            The remote node is expected to reply with a Message with the :const:`~neo3.network.message.MessageType.INV`
            type containing the hashes of the requested blocks. Use :meth:`~neo3.network.node.NeoNode.request_data` in
            combination with these hashes to return the actual :class:`~neo3.network.payloads.block.Block` objects.

        See also:
            :meth:`~neo3.network.node.NeoNode.request_block_data()` to immediately retrieve
            :class:`~neo3.network.payloads.block.Block` objects.

        Args:
            hash_start:
            count:
        """
        m = message.Message(msg_type=message.MessageType.GETBLOCKS,
                            payload=payloads.GetBlocksPayload(hash_start, count))
        await self.send_message(m)

    async def request_block_data(self, index_start, count) -> None:
        """
        Send a request for `count` blocks starting from `index_start`.

        Count cannot exceed :attr:`~neo3.network.payloads.block.GetBlockByIndexPayload.MAX_BLOCKS_COUNT`.

        See also:
            :meth:`~neo3.network.node.NeoNode.request_blocks()` to only request block hashes.

        Args:
            index_start: block index to start from.
            count: number of blocks to return.
        """
        m = message.Message(msg_type=message.MessageType.GETBLOCKBYINDEX,
                            payload=payloads.GetBlockByIndexPayload(index_start, count))
        await self.send_message(m)

    async def request_data(self, type: payloads.InventoryType, hashes: List[types.UInt256]) -> None:
        """
        Send a request for receiving the specified inventory data.

        Args:
            type:
            hashes:
        """
        if len(hashes) < 1:
            return

        m = message.Message(msg_type=message.MessageType.GETDATA, payload=payloads.InventoryPayload(type, hashes))
        await self.send_message(m)

    async def send_inventory(self, inv_type: payloads.InventoryType, inv_hash: types.UInt256) -> None:
        """
        Send an inventory to the network

        Args:
            inv_type:
            inv_hash:
        """
        inv = payloads.InventoryPayload(type=inv_type, hashes=[inv_hash])
        m = message.Message(msg_type=message.MessageType.INV, payload=inv)
        await self.send_message(m)

    async def send_ping(self) -> None:
        """
        Send a Ping message and expecting a Pong response
        """
        if settings.database:
            height = max(0, blockchain.Blockchain().height)
        else:
            height = 0

        ping = payloads.PingPayload(height)
        m = message.Message(msg_type=message.MessageType.PING, payload=ping)
        await self.send_message(m)

    async def relay(self, inventory: payloads.IInventory) -> bool:
        """
        Relay the inventory to the network

        Args:
            inventory: should be of type Block, Transaction or Consensus. See: :class:`~neo3.network.payloads.inventory.InventoryType`. # noqa
        """
        relaycache.RelayCache().add(inventory)
        await self.send_inventory(inventory.inventory_type, inventory.hash())
        return True

    # utility functions
    async def send_message(self, message: message.Message) -> None:
        """
        Send a Message over the wire.

        Args:
            message:
        """
        await self.protocol.send_message(message)

    async def read_message(self, timeout: int = 30) -> Optional[message.Message]:
        """
        Read a Message from the wire.

        Args:
            timeout: maximum time to wait in trying to deserialize a message from the wire.

        Returns:
            Message, if enough data was found and successfully deserialized.
            None otherwise.
        """
        return await self.protocol.read_message(timeout)

    def __eq__(self, other):
        if type(other) is type(self):
            return self.address == other.address and self.nodeid == other.nodeid
        else:
            return False

    def __repr__(self):
        return f"<{self.__class__.__name__} at {hex(id(self))}> {self.nodeid_human}"

    @staticmethod
    async def connect_to(host: str = None,
                         port: int = None,
                         timeout=3,
                         loop=None,
                         socket=None) -> Tuple[Optional[NeoNode], Optional[Tuple[str, str]]]:
        """
        Establish a connection to a Neo node

        Note: performs the initial connection handshake and validation.

        Args:
            host: remote address in IPv4 format
            port: remote port
            timeout: maximum time establishing a connection may take
            loop: custom loop

        Raises:
            ValueError: if host/port and the socket argument as specified as the same time or none are specified.

        Returns:
            Tuple:
                - (Node instance, None) - if a connection was successfully established
                - (None, (ip address, error reason)) - if a connection failed to establish . Reasons include connection timeout, connection full and handshake errors. # noqa
        """
        if loop is None:
            loop = asyncio.get_event_loop()

        if host is not None or port is not None:
            if socket is not None:
                raise ValueError('host/port and socket can not be specified at the same time')
        if socket is None and (host is None or port is None):
            raise ValueError('host and port was not specified and no sock specified')

        try:
            if socket:
                logger.debug(f"Trying to connect to socket: {socket}.")
                connect_coro = loop.create_connection(protocol.NeoProtocol, sock=socket)
            else:
                logger.debug(f"Trying to connect to: {host}:{port}.")
                connect_coro = loop.create_connection(protocol.NeoProtocol, host, port, family=IP4_FAMILY)
            transport, node = await asyncio.wait_for(connect_coro, timeout)

            success, fail_reason = await node.client._do_handshake()
            if success:
                return node.client, None
            else:
                raise Exception(fail_reason)
        except asyncio.TimeoutError:
            reason = f"Timed out"
        except OSError as e:
            reason = f"Failed to connect for reason {e}"
        except asyncio.CancelledError:
            reason = "Cancelled"
        except Exception as e:
            reason = traceback.format_exc()
        return None, (f"{host}:{port}", reason)

    @classmethod
    def get_address_new(cls) -> Optional[payloads.NetworkAddress]:
        """
        Utility function to return the first address with the state NEW.
        """
        for addr in cls.addresses:
            if addr.is_state_new:
                return addr
        # explicit return to silence mypy
        return None

    def _find_address_by_host_port(self, host_port) -> Optional[payloads.NetworkAddress]:
        addr = payloads.NetworkAddress(address=host_port)
        try:
            idx = self.addresses.index(addr)
            return self.addresses[idx]
        except ValueError:
            return None

    def _create_task_with_cleanup(self, coro):
        task = asyncio.create_task(coro)
        self.tasks.append(task)
        task.add_done_callback(lambda fut: self.tasks.remove(fut))

    def start_message_handler(self) -> None:
        """
        A convenience function to start a message reading loop and forward the messages to their respective handlers as
        configured in :attr:`~neo3.network.node.NeoNode.dispatch_table`.
        """
        # when we break out of the read/write loops, we should make sure we disconnect
        self._read_task = asyncio.create_task(self._process_incoming_data())
        self._read_task.add_done_callback(lambda _:
                                          asyncio.create_task(self.disconnect(payloads.DisconnectReason.UNKNOWN)))

    @classmethod
    def _reset_for_test(cls) -> None:
        cls.addresses = []



import unittest
from neo3.core import jsonpath

json_data = {
    "store": {
        "book": [
            {
                "category": "reference",
                "author": "Nigel Rees",
                "title": "Sayings of the Century",
                "price": 8.95
            },
            {
                "category": "fiction",
                "author": "Evelyn Waugh",
                "title": "Sword of Honour",
                "price": 12.99
            },
            {
                "category": "fiction",
                "author": "Herman Melville",
                "title": "Moby Dick",
                "isbn": "0-553-21311-3",
                "price": 8.99
            },
            {
                "category": "fiction",
                "author": "J. R. R. Tolkien",
                "title": "The Lord of the Rings",
                "isbn": "0-395-19395-8",
                "price": 22.99
            }
        ],
        "bicycle": {
            "color": "red",
            "price": 19.95
        },
    },
    "expensive": 10,
}


class JsonPathTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.json = jsonpath.JsonPath(json_data)

    def test_csharp_cases(self):
        self.assertEqual(['Nigel Rees', 'Evelyn Waugh', 'Herman Melville', 'J. R. R. Tolkien'], self.json.parse("$.store.book[*].author"))
        self.assertEqual(['Nigel Rees', 'Evelyn Waugh', 'Herman Melville', 'J. R. R. Tolkien'], self.json.parse("$..author"))
        self.assertEqual([19.95,8.95,12.99,8.99,22.99], self.json.parse("$.store..price"))
        self.assertEqual([{'category': 'fiction', 'author': 'Herman Melville', 'title': 'Moby Dick', 'isbn': '0-553-21311-3', 'price': 8.99}], self.json.parse("$..book[2]"))
        self.assertEqual([{'category': 'fiction', 'author': 'Herman Melville', 'title': 'Moby Dick', 'isbn': '0-553-21311-3', 'price': 8.99}], self.json.parse("$..book[-2]"))
        self.assertEqual([{'category': 'reference', 'author': 'Nigel Rees', 'title': 'Sayings of the Century', 'price': 8.95}, {'category': 'fiction', 'author': 'Evelyn Waugh', 'title': 'Sword of Honour', 'price': 12.99}], self.json.parse("$..book[0,1]"))
        self.assertEqual([{'category': 'reference', 'author': 'Nigel Rees', 'title': 'Sayings of the Century', 'price': 8.95}, {'category': 'fiction', 'author': 'Evelyn Waugh', 'title': 'Sword of Honour', 'price': 12.99}], self.json.parse("$..book[:2]"))
        self.assertEqual([{'category': 'fiction', 'author': 'Evelyn Waugh', 'title': 'Sword of Honour', 'price': 12.99}], self.json.parse("$..book[1:2]"))
        self.assertEqual([{'category': 'fiction', 'author': 'Herman Melville', 'title': 'Moby Dick', 'isbn': '0-553-21311-3', 'price': 8.99}, {'category': 'fiction', 'author': 'J. R. R. Tolkien', 'title': 'The Lord of the Rings', 'isbn': '0-395-19395-8', 'price': 22.99}], self.json.parse("$..book[-2:]"))
        self.assertEqual([{'category': 'fiction', 'author': 'Herman Melville', 'title': 'Moby Dick', 'isbn': '0-553-21311-3', 'price': 8.99}, {'category': 'fiction', 'author': 'J. R. R. Tolkien', 'title': 'The Lord of the Rings', 'isbn': '0-395-19395-8', 'price': 22.99}], self.json.parse("$..book[2:]"))
        self.assertEqual([{'book': [{'category': 'reference', 'author': 'Nigel Rees', 'title': 'Sayings of the Century', 'price': 8.95}, {'category': 'fiction', 'author': 'Evelyn Waugh', 'title': 'Sword of Honour', 'price': 12.99}, {'category': 'fiction', 'author': 'Herman Melville', 'title': 'Moby Dick', 'isbn': '0-553-21311-3', 'price': 8.99}, {'category': 'fiction', 'author': 'J. R. R. Tolkien', 'title': 'The Lord of the Rings', 'isbn': '0-395-19395-8', 'price': 22.99}], 'bicycle': {'color': 'red', 'price': 19.95}}, 10], self.json.parse("$.*"))






from __future__ import annotations
from typing import List, Optional
from datetime import datetime
from netaddr import IPAddress  # type: ignore
from enum import IntEnum
from neo3.core import serialization, Size as s, utils
from neo3.network.payloads import VersionPayload
from neo3.network import capabilities


class AddressState(IntEnum):
    """
    Node state

    Used for tracking remote address state.
    """
    #: an address that has not been used before in the current connection cycle.
    NEW = 0x0

    #: an address that is associated with a live node and performing OK.
    CONNECTED = 0x1

    #: an address from a node that has been disconnected from due to
    #:
    #: - bad performance
    #: - malformed data or protocol responses
    #: - or max connections reached reasons.
    POOR = 0x2

    #: an address that could not be resolved or connected to due to a timeout.
    DEAD = 0x3


class DisconnectReason(IntEnum):
    """
    Reason for disconnecting a node.

    Will also be broadcasted back to the node when `this PR <https://github.com/neo-project/neo/pull/1154>`__ is merged.
    For now only used internally with logging.
    """
    UNKNOWN = 0x0
    MAX_CONNECTIONS_REACHED = 0x1
    POOR_PERFORMANCE = 0x2
    IPFILTER_NOT_ALLOWED = 0x3
    HANDSHAKE_VERSION_ERROR = 0x4
    HANDSHAKE_VERACK_ERROR = 0x5
    SHUTTING_DOWN = 0x6


class NetworkAddress(serialization.ISerializable):

    def __init__(self,
                 address: str,
                 capabilities: List[capabilities.NodeCapability] = None,
                 timestamp: int = None,
                 state: AddressState = AddressState.NEW):
        """ Create an instance. """
        if timestamp is None:
            self.timestamp = int(datetime.utcnow().timestamp())
        else:
            self.timestamp = timestamp

        self.address = address if address else '0.0.0.0:0'  # host:port
        self.capabilities = capabilities if capabilities else []

        # none official properties
        self.state = state
        self.disconnect_reason = None  # type: Optional[DisconnectReason]
        self.last_connected = 0

    def __len__(self):
        """
        Get the total size in bytes of the object.

        Note:
            This size is only for the official properties such that it matches C#.
            Unofficial properties are not accounted for.
        """
        return s.uint32 + 16 + utils.get_var_size(self.capabilities)

    def __eq__(self, other):
        if type(other) is type(self):
            return self.address == other.address
        else:
            return False

    def __repr__(self):
        return f"<{self.__class__.__name__} at {hex(id(self))}> {self.address} ({self.state.name})"

    def __str__(self):
        return self.address

    def __format__(self, format_spec):
        return self.address.__format__(format_spec)

    def __hash__(self):
        return hash((self.address, self.timestamp))

    @property
    def ip(self) -> str:
        host, port = self.address.split(":")
        return host

    @property
    def port(self) -> int:
        host, port = self.address.split(":")
        return int(port)

    @property
    def is_state_new(self) -> bool:
        return self.state == AddressState.NEW

    def set_state_new(self) -> None:
        self.state = AddressState.NEW

    @property
    def is_state_connected(self) -> bool:
        return self.state == AddressState.CONNECTED

    def set_state_connected(self) -> None:
        self.state = AddressState.CONNECTED

    @property
    def is_state_poor(self) -> bool:
        return self.state == AddressState.POOR

    def set_state_poor(self) -> None:
        self.state = AddressState.POOR

    @property
    def is_state_dead(self) -> bool:
        return self.state == AddressState.DEAD

    def set_state_dead(self) -> None:
        self.state = AddressState.DEAD

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_uint32(self.timestamp)
        ip = IPAddress(self.ip).ipv6()
        writer.write_bytes(ip.packed)
        writer.write_serializable_list(self.capabilities)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.timestamp = reader.read_uint32()

        full_address_bytes = bytearray(reader.read_bytes(16))
        ip_bytes = full_address_bytes[-4:]
        host = '.'.join(map(lambda b: str(b), ip_bytes))
        port = 0

        capabilities_cnt = reader.read_var_int(VersionPayload.MAX_CAPABILITIES)
        capabilities_list = []
        for _ in range(capabilities_cnt):
            capa = capabilities.NodeCapability.deserialize_from(reader)
            if isinstance(capa, capabilities.ServerCapability):
                port = capa.port
            capabilities_list.append(capa)
        self.capabilities = capabilities_list
        self.address = f"{host}:{port}"

    @classmethod
    def _serializable_init(cls):
        return cls("0.0.0.0:0")


class AddrPayload(serialization.ISerializable):
    def __init__(self, addresses: List[NetworkAddress]):
        self.addresses = addresses

    def __len__(self):
        return utils.get_var_size(self.addresses)

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        writer.write_var_int(len(self.addresses))
        for address in self.addresses:
            address.serialize(writer)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        addr_list_len = reader.read_var_int()
        for i in range(0, addr_list_len):
            nawt = NetworkAddress("0.0.0.0:0")
            nawt.deserialize(reader)
            self.addresses.append(nawt)

    @classmethod
    def _serializable_init(cls):
        return cls([])



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
to_add = os.path.abspath('../../neo3')
# print(f"Path adding! {to_add}")
sys.path.insert(0, to_add)
sys.path.append(os.path.abspath("./_ext"))
sys.path.append(os.path.abspath("./_theme"))
from neo3 import version


# -- Project information -----------------------------------------------------

project = 'neo-mamba'
copyright = '2019-2021, COZ - Erik van den Brink'
author = 'Erik van den Brink'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
'sphinx.ext.autodoc',
'sphinx.ext.napoleon',
'sphinx_autodoc_typehints',
'sphinx.ext.intersphinx',
'classoverview'
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

# use a modified 'classic' theme, can be found in the _theme directory.
html_theme = 'neo3'
# provide our own landing page
html_additional_pages = {'index': 'index.html'}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# napoleon specific configuration
napoleon_numpy_docstring = False
# don't include dunder methods
napoleon_include_special_with_doc = False
# must be set to True or param types get stripped from the documentation
napoleon_use_param = True

# autodoc specific configuration
# set typing.TYPE_CHECKING to True to enable “expensive” typing imports
set_type_checking_flag = True
autoclass_content = 'both'
autodoc_member_order = 'groupwise'

# intersphinx configuratin
intersphinx_mapping = {'python': ('https://docs.python.org/3', None)}



# type: ignore

from enum import Enum
from collections import Iterable
from neo3.core import serialization
from neo3.core import Size


def get_var_size(value: object) -> int:
    """
    Determine the variable size of an object.

    Note:
        This function is only used internally for correctly returning object sizes of network payloads to be in line
        with C#'s output.

    Args:
        value: any object type

    Raises:
        TypeError: if a specific Iterable type is not supported.
        ValueError: if a specific object type is not supported .
    """
    # public static int GetVarSize(this string value)
    if isinstance(value, str):
        value_size = len(value.encode('utf-8'))
        return get_var_size(value_size) + value_size

    # internal static int GetVarSize(int value)
    elif isinstance(value, int):
        if (value < 0xFD):
            return Size.uint8
        elif (value <= 0xFFFF):
            return Size.uint8 + Size.uint16
        else:
            return Size.uint8 + Size.uint32

    # internal static int GetVarSize<T>(this T[] value)
    elif isinstance(value, Iterable):
        value_length = len(value)
        value_size = 0

        if value_length > 0:
            if isinstance(value[0], serialization.ISerializable):
                value_size = sum(map(lambda t: len(t), value))
            elif isinstance(value[0], Enum):
                # Note: currently all Enum's in neo core (C#) are of type Byte. Only porting that part of the code
                value_size = value_length * Size.uint8
            elif isinstance(value, (bytes, bytearray)):
                # experimental replacement for: value_size = value.Length * Marshal.SizeOf<T>();
                # because I don't think we have a reliable 'SizeOf' in python
                value_size = value_length * Size.uint8
            else:
                raise TypeError(
                    f"Cannot accurately determine size of objects that do not inherit from 'ISerializable', "
                    f"'Enum' or 'bytes'. Found type: {type(value[0])}")
    else:
        raise ValueError(f"[NOT SUPPORTED] Unexpected value type {type(value)} for get_var_size()")

    return get_var_size(value_length) + value_size



from __future__ import annotations
import lz4.block  # type: ignore
from enum import IntEnum, IntFlag
from neo3.network import payloads
from neo3.core import Size as s, serialization
from neo3 import network_logger as logger


class MessageType(IntEnum):
    VERSION = 0x00
    VERACK = 0x01

    GETADDR = 0x10
    ADDR = 0x11
    PING = 0x18
    PONG = 0x19

    GETHEADERS = 0x20
    HEADERS = 0x21
    GETBLOCKS = 0x24
    MEMPOOL = 0x25
    INV = 0x27
    GETDATA = 0x28
    GETBLOCKBYINDEX = 0x29
    NOTFOUND = 0x2a
    TRANSACTION = 0x2b
    BLOCK = 0x2c
    CONSENSUS = 0x2d
    EXTENSIBLE = 0x2e
    REJECT = 0x2f

    FILTERLOAD = 0x30
    FILTERADD = 0x31
    FILTERCLEAR = 0x32
    MERKLEBLOCK = 0x38

    ALERT = 0x40

    DEFAULT = 0xFF  # not supported in the real protocol


class MessageConfig(IntFlag):
    #: Indicates that the payload data is not compressed
    NONE = 0
    #: Indicates that the payload data is compressed using LZ4
    COMPRESSED = 1 << 0


class Message(serialization.ISerializable):
    PAYLOAD_MAX_SIZE = 0x2000000
    COMPRESSION_MIN_SIZE = 128
    COMPRESSION_THRESHOLD = 64

    def __init__(self, msg_type: MessageType, payload: serialization.ISerializable_T = None):
        """

        Args:
            msg_type: message object configuration.
            payload: an identifier specifying the purpose of the message.
        """
        self.config = MessageConfig.NONE  #: MessageConfig: message object configuration.
        #: MessageType: an identifier specifying the purpose of the message.
        self.type: MessageType = msg_type
        self.payload: serialization.ISerializable_T = payloads.EmptyPayload()  # type: ignore
        # mypy doesn't get EmptyPayload is an ISerializable

        if payload:
            self.payload = payload

    def __len__(self):
        """ Get the total size in bytes of the object. """
        return s.uint8 + s.uint8 + len(self.payload)

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        """
        Serialize the object into a binary stream.

        Args:
            writer: instance.
        """
        payload = self.payload.to_array()

        if len(self.payload) > self.COMPRESSION_MIN_SIZE and MessageConfig.COMPRESSED not in self.config:
            compressed_data = lz4.block.compress(self.payload.to_array(), store_size=False)
            compressed_data = len(payload).to_bytes(4, 'little') + compressed_data
            if len(compressed_data) < len(self.payload) - self.COMPRESSION_THRESHOLD:
                payload = compressed_data
                self.config |= MessageConfig.COMPRESSED

        writer.write_uint8(self.config)
        writer.write_uint8(self.type.value)
        writer.write_var_bytes(payload)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        """
        Deserialize the object from a binary stream.

        Args:
            reader: instance.
        """
        self.config = MessageConfig(reader.read_uint8())
        x = reader.read_uint8()
        self.type = MessageType(x)
        # self.type = MessageType(reader.read_uint8())

        payload_data = reader.read_var_bytes(self.PAYLOAD_MAX_SIZE)
        if len(payload_data) > 0:
            if MessageConfig.COMPRESSED in self.config:
                # From the lz4 documentation:
                # "The uncompressed_size argument specifies an upper bound on the size of the uncompressed data size
                # rather than an absolute value"
                try:
                    size = int.from_bytes(payload_data[:4], 'little')
                    payload_data = lz4.block.decompress(payload_data[4:], uncompressed_size=size)
                except lz4.block.LZ4BlockError:
                    raise ValueError("Invalid payload data - decompress failed")

            self.payload = self._payload_from_data(self.type, payload_data)

        if self.payload is None:
            self.payload = payloads.EmptyPayload()

    @staticmethod
    def _payload_from_data(msg_type, data):
        with serialization.BinaryReader(data) as br:
            if msg_type in [MessageType.INV, MessageType.GETDATA]:
                return br.read_serializable(payloads.InventoryPayload)
            elif msg_type == MessageType.GETBLOCKBYINDEX:
                return br.read_serializable(payloads.GetBlockByIndexPayload)
            elif msg_type == MessageType.VERSION:
                return br.read_serializable(payloads.VersionPayload)
            elif msg_type == MessageType.VERACK:
                return br.read_serializable(payloads.EmptyPayload)
            elif msg_type == MessageType.BLOCK:
                return br.read_serializable(payloads.Block)
            elif msg_type == MessageType.HEADERS:
                return br.read_serializable(payloads.HeadersPayload)
            elif msg_type in [MessageType.PING, MessageType.PONG]:
                return br.read_serializable(payloads.PingPayload)
            elif msg_type == MessageType.ADDR:
                return br.read_serializable(payloads.AddrPayload)
            elif msg_type == MessageType.TRANSACTION:
                return br.read_serializable(payloads.Transaction)
            elif msg_type == MessageType.EXTENSIBLE:
                return br.read_serializable(payloads.ExtensiblePayload)
            else:
                logger.debug(f"Unsupported payload {msg_type.name}")

    @classmethod
    def _serializable_init(cls):
        return cls(MessageType.DEFAULT)



from docutils import nodes
from sphinx.util.docutils import SphinxDirective


class ClassOverviewNode(nodes.General, nodes.Element):
    pass


class ClassOverview(SphinxDirective):
    """
    Add a .. classoverview:: directive
    through which we can filter the sidebar with a list of classes discussed in the module
    """
    has_content = True

    def run(self):
        if not hasattr(self.env, 'side_bar_classes'):
            self.env.side_bar_classes = {}

        co_node = ClassOverviewNode()
        co_node._neo3 = self.env.docname

        # this seems to trigger building of partial_xref's, that are resolved by the time
        # the "doctree-resolved" callback is called. Doesn't work without it.
        self.state.nested_parse(self.content, self.content_offset, co_node)

        # we add the node for the time being so we can easily find it when we do the transform step below
        return [co_node]


def transform(app, doctree, fromdocname):
    document_class_links = app.env.side_bar_classes.get(fromdocname, None)

    for n in doctree.traverse(ClassOverviewNode):
        if hasattr(n, '_neo3') and n._neo3 == fromdocname:
            if document_class_links is None:
                app.env.side_bar_classes[fromdocname] = []
            for paragraph in n.children:
                for child in paragraph.children:
                    if isinstance(child, nodes.reference):
                        app.env.side_bar_classes[fromdocname].append(child.attributes['refid'])
            # we don't want to keep the node in the doc that has the directive because it will be moved to the sidebar
            n.parent.remove(n)


def update_context(app, pagename, templatename, context, doctree):
    # Here we update the context passed to our classoverview.html template to include our data
    #
    # This is called before the template is rendered, but after the initial context is filled
    try:
        # for some reason alabaster doesn't have this property, but is part of the default installation.
        docname = context['docname']
    except KeyError:
        return

    # for some unknown reason the dictionary passed to the template cannot be accessed by key inside the template
    # it always throws "UndefinedError" while iterating over the dictionary does work, but that's useless for this use-case
    # we'll just build a list instead which can be accessed by index ¯\_(ツ)_/¯
    list_refids = app.env.side_bar_classes.get(context['docname'], None)
    if list_refids:
        pairs = []
        for refid in list_refids:
            parts = refid.split('.')
            name = parts[-1]
            pairs.append([name, refid])
        pairs.sort(key=lambda pair: pair[0])
        if len(pairs) > 0:
            context['classoverview_links'] = pairs
        else:
            context['classoverview_links'] = None


def purge_classoverview_links(app, env, docname):
    if not hasattr(env, 'side_bar_classes'):
        return

    env.side_bar_classes = {}


def show_all_attrs(value):
    res = []
    for k in dir(value):
        res.append('%r %r\n' % (k, getattr(value, k)))
    return '\n'.join(res)


def add_jinja_filters(app):
    app.builder.templates.environment.filters['dir'] = show_all_attrs


def setup(app):
    app.add_node(ClassOverviewNode),
    app.add_directive('classoverview', ClassOverview)
    app.connect('doctree-resolved', transform)
    app.connect('html-page-context', update_context)
    app.connect('env-purge-doc', purge_classoverview_links)

    # remove after debug
    # app.connect('builder-inited', add_jinja_filters)

    return {
        'version': '0.1',
        'parallel_read_safe': False,
        'parallel_write_safe': False,
    }


import unittest
import base64
from typing import Optional, Any
from aioresponses import aioresponses
from neo3 import api
from neo3.network import payloads
from neo3.core import types, cryptography

JSON = Any


class TestNeoRpcClient(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.client = api.NeoRpcClient('localhost')
        self.helper = aioresponses()
        self.helper.start()

    async def asyncTearDown(self) -> None:
        self.helper.stop()
        await self.client.close()

    def mock_response(self, payload: Optional[JSON] = None, exc: Optional[Exception] = None):
        """
        Either payload or exc should be provided
        """
        if payload is not None and exc is not None:
            raise ValueError("Arguments are mutual exclusive")

        if payload is not None:
            json = {'jsonrpc': '2.0', 'id': 1, "result": payload}
            self.helper.post('localhost', payload=json)
        else:
            self.helper.post('localhost', exception=exc)

    async def test_calculate_network_fee(self):
        self.mock_response({"networkfee": "123"})
        response = await self.client.calculate_network_fee(payloads.Transaction._serializable_init())
        self.assertEqual(123, response)

    async def test_get_application_log(self):
        captured = {
            "txid": "0x7da6ae7ff9d0b7af3d32f3a2feb2aa96c2a27ef8b651f9a132cfaad6ef20724c",
            "executions": [
                {
                    "trigger": "Application",
                    "vmstate": "HALT",
                    "exception": None,
                    "gasconsumed": "9999540",
                    "stack": [],
                    "notifications": [
                        {
                            "contract": "0x70e2301955bf1e74cbb31d18c2f96972abadb328",
                            "eventname": "Transfer",
                            "state": {
                                "type": "Array",
                                "value": [
                                    {
                                        "type": "ByteString",
                                        "value": "4rZTInKT6ZxPKQbVNVOrtKZy34Y="
                                    },
                                    {
                                        "type": "ByteString",
                                        "value": "+on7LBTfD1nd3wT25WUX8rNKrus="
                                    },
                                    {
                                        "type": "Integer",
                                        "value": "10000000000"
                                    }
                                ]
                            }
                        }
                    ]
                }
            ]
        }
        self.mock_response(captured)
        tx_hash = types.UInt256.from_string("0x7da6ae7ff9d0b7af3d32f3a2feb2aa96c2a27ef8b651f9a132cfaad6ef20724c"[2:])
        response = await self.client.get_application_log(tx_hash)
        self.assertEqual(1, len(response.executions))
        self.assertEqual(0, len(response.executions[0].stack))
        self.assertEqual("HALT", response.executions[0].state)
        self.assertEqual(1, len(response.executions[0].notifications))
        self.assertEqual("Transfer", response.executions[0].notifications[0].event_name)

    async def test_get_best_block_hash(self):
        hash_ = "0xbee7a65279d6b31cc45445a7579d4c4a4e52d1edc13cc7ec7a41f7b1affdf0ab"
        self.mock_response(hash_)
        response = await self.client.get_best_block_hash()
        self.assertEqual(types.UInt256.from_string(hash_[2:]), response)

    async def test_get_block(self):
        block = payloads.Block._serializable_init()
        self.mock_response(base64.b64encode(block.to_array()).decode())
        response_block = await self.client.get_block(1)
        self.assertEqual(block, response_block)

    async def test_get_block_count(self):
        count = 1
        self.mock_response(count)
        response = await self.client.get_block_count()
        self.assertEqual(count, response)

    async def test_get_block_hash(self):
        hash_ = "0xbee7a65279d6b31cc45445a7579d4c4a4e52d1edc13cc7ec7a41f7b1affdf0ab"
        self.mock_response(hash_)
        response = await self.client.get_block_hash(1)
        self.assertEqual(types.UInt256.from_string(hash_[2:]), response)

    async def test_get_block_header(self):
        block = payloads.Block._serializable_init()
        self.mock_response(base64.b64encode(block.to_array()).decode())
        response_header = await self.client.get_block_header(1)
        self.assertEqual(block.header, response_header)

    async def test_get_committee(self):
        points = [
            "02237309a0633ff930d51856db01d17c829a5b2e5cc2638e9c03b4cfa8e9c9f971",
            "022f1beae94cf0d266d7d26691b431958c8d13768103ab20aed817b57568da293f",
            "0239a37436652f41b3b802ca44cbcb7d65d3aa0b88c9a0380243bdbe1aaa5cb35b"
        ]
        self.mock_response(points)
        expected = list(map(lambda p: cryptography.ECPoint.deserialize_from_bytes(bytes.fromhex(p)), points))
        committee = await self.client.get_committee()
        self.assertEqual(expected, committee)

    async def test_get_connection_count(self):
        count = 1
        self.mock_response(count)
        response = await self.client.get_connection_count()
        self.assertEqual(count, response)

    async def test_get_contract_state(self):
        contract_hash_str = "0xb776afb6ad0c11565e70f8ee1dd898da43e51be1"
        contract_hash = types.UInt160.from_string(contract_hash_str[2:])
        captured = {'id': 1, 'updatecounter': 0, 'hash': contract_hash_str, 'nef': {'magic': 860243278, 'compiler': 'Neo.Compiler.CSharp 3.0.0', 'source': '', 'tokens': [{'hash': '0xfffdc93764dbaddd97c48f252a53ea4643faa3fd', 'method': 'update', 'paramcount': 3, 'hasreturnvalue': False, 'callflags': 'All'}, {'hash': '0xef4073a0f2b305a38ec4050e4d3d28bc40ea63f5', 'method': 'getCandidates', 'paramcount': 0, 'hasreturnvalue': True, 'callflags': 'All'}, {'hash': '0x726cb6e0cd8628a1350a611384688911ab75f51b', 'method': 'ripemd160', 'paramcount': 1, 'hasreturnvalue': True, 'callflags': 'All'}, {'hash': '0x726cb6e0cd8628a1350a611384688911ab75f51b', 'method': 'sha256', 'paramcount': 1, 'hasreturnvalue': True, 'callflags': 'All'}, {'hash': '0xacce6fd80d44e1796aa0c2c625e9e4e0ce39efc0', 'method': 'serialize', 'paramcount': 1, 'hasreturnvalue': True, 'callflags': 'All'}, {'hash': '0xacce6fd80d44e1796aa0c2c625e9e4e0ce39efc0', 'method': 'deserialize', 'paramcount': 1, 'hasreturnvalue': True, 'callflags': 'All'}], 'script': 'NA5B+CfsjEBB+CfsjEBXAQAMCnN1cGVyQWRtaW40IHBoStgkA8oAFJcmEGhK2CQJSsoAFCgDOiIDWCICQFcAAXhBm/ZnzkGSXegxIgJAQZJd6DFAQZv2Z85AykBXAAEMCUZvcmJpZGRlbjSkQfgn7Iw0F3gMCnN1cGVyQWRtaW40EhHbICICQFcAAniqJgR5OkBXAAJ5eEGb9mfOQeY/GIRAQeY/GIRAVwADNVX///+qJhYMEU5vIGF1dGhvcml6YXRpb24uOnp5eDcAAEA3AABAVwABWXjbMItAi0DbMEBXCAoMCUZvcmJpZGRlbnhB+CfsjDSUNwEAcBDbIHFoSnLKcxB0Ih9qbM7BRXV2bTRhdwd4bweXJgoR2yBKcUUiCWycdGxrMOEMF1NlbmRlciBpcyBub3QgQ2FuZGlkYXRlaTVG////fwl/CH8Hfn18e3p5eBrASjRLcmo3BAB4NXP///80QhHbICICQDcBAEBXAgFaeNswi1uLcGjbKDcDADcCAHFpStgkCUrKABQoAzoiAkDbMEA3AgBANwMAQNsoQFcAAUBXAAJ5eEGb9mfOQeY/GIRAQeY/GIRANwQAQFcBAXg1Dv///zQXcGhK2CQDyhC3JghoNwUAIgULIgJAVwABeEGb9mfOQZJd6DEiAkBBkl3oMUA3BQBAVwMAWTQocBDEAHFoQZwI7ZwmF2hB81S/HXJqEc4LmCYHaWoRzs8i5WkiAkBXAAEaeEGb9mfOQd8wuJoiAkBB3zC4mkBBnAjtnEBB81S/HUDPQFcCAQwJRm9yYmlkZGVueEH4J+yMJgcR2yAiBzWY/f//NRv+//94NV/+//9weDVY/v//NWH///9xaUrYJAPKELcmCmg0DRHbICIHENsgIgJAVwABeNsoQZv2Z85BL1jF7UBBL1jF7UDPQFYEDBTARUMMYSJWDL3Fhow6TOAvAt28wWAMAgwh2zBiDAVBVuezJ9swYwwBd9swYUA=', 'checksum': 3038242458}, 'manifest': {'name': 'CommitteeInfoContract', 'groups': [], 'features': {}, 'supportedstandards': [], 'abi': {'methods': [{'name': 'verify', 'parameters': [], 'returntype': 'Boolean', 'offset': 0, 'safe': False}, {'name': 'getAdmin', 'parameters': [], 'returntype': 'Hash160', 'offset': 14, 'safe': False}, {'name': 'setAdmin', 'parameters': [{'name': 'admin', 'type': 'Hash160'}], 'returntype': 'Boolean', 'offset': 92, 'safe': False}, {'name': 'update', 'parameters': [{'name': 'nefFile', 'type': 'ByteArray'}, {'name': 'manifest', 'type': 'String'}, {'name': 'data', 'type': 'Any'}], 'returntype': 'Void', 'offset': 168, 'safe': False}, {'name': 'setInfo', 'parameters': [{'name': 'sender', 'type': 'Hash160'}, {'name': 'name', 'type': 'String'}, {'name': 'location', 'type': 'String'}, {'name': 'website', 'type': 'String'}, {'name': 'email', 'type': 'String'}, {'name': 'github', 'type': 'String'}, {'name': 'telegram', 'type': 'String'}, {'name': 'twitter', 'type': 'String'}, {'name': 'description', 'type': 'String'}, {'name': 'logo', 'type': 'String'}], 'returntype': 'Boolean', 'offset': 224, 'safe': False}, {'name': 'getInfo', 'parameters': [{'name': 'candidate', 'type': 'Hash160'}], 'returntype': 'Any', 'offset': 448, 'safe': False}, {'name': 'getAllInfo', 'parameters': [], 'returntype': 'Array', 'offset': 507, 'safe': False}, {'name': 'deleteInfo', 'parameters': [{'name': 'candidate', 'type': 'Hash160'}], 'returntype': 'Boolean', 'offset': 589, 'safe': False}, {'name': '_initialize', 'parameters': [], 'returntype': 'Void', 'offset': 694, 'safe': False}], 'events': []}, 'permissions': [{'contract': '0x726cb6e0cd8628a1350a611384688911ab75f51b', 'methods': ['ripemd160', 'sha256']}, {'contract': '0xacce6fd80d44e1796aa0c2c625e9e4e0ce39efc0', 'methods': ['deserialize', 'serialize']}, {'contract': '0xef4073a0f2b305a38ec4050e4d3d28bc40ea63f5', 'methods': ['getCandidates']}, {'contract': '0xfffdc93764dbaddd97c48f252a53ea4643faa3fd', 'methods': ['update']}], 'trusts': [], 'extra': {'Author': 'NEO', 'Email': 'developer@neo.org', 'Description': 'This is a Neo3 Contract'}}}
        self.mock_response(captured)

        response_state = await self.client.get_contract_state(contract_hash)
        self.assertEqual(contract_hash, response_state.hash)
        self.assertEqual("CommitteeInfoContract", response_state.manifest.name)

    async def test_get_nep17_balances(self):
        captured = {
            "balance": [
                {
                    "assethash": "0x70e2301955bf1e74cbb31d18c2f96972abadb328",
                    "amount": "3000000100000000",
                    "lastupdatedblock": 2
                },
                {
                    "assethash": "0xf61eebf573ea36593fd43aa150c055ad7906ab83",
                    "amount": "99999900",
                    "lastupdatedblock": 2
                }
            ],
            "address": "NgaiKFjurmNmiRzDRQGs44yzByXuSkdGPF"
        }
        self.mock_response(captured)

        response = await self.client.get_nep17_balances("bogus_addr")
        self.assertEqual(2, len(response.balances))
        self.assertEqual(99999900, response.balances[1].amount)

    async def test_get_nep17_transfers(self):
        captured = {
            "sent": [],
            "received": [
                {
                    "timestamp": 1612690497725,
                    "assethash": "0xf61eebf573ea36593fd43aa150c055ad7906ab83",
                    "transferaddress": "NgaiKFjurmNmiRzDRQGs44yzByXuSkdGPF",
                    "amount": "100",
                    "blockindex": 2,
                    "transfernotifyindex": 1,
                    "txhash": "0x5f957960a782514d6587c445288ee1cca7d6b0f952edc204f14d9be83b8152ff"
                },
                {
                    "timestamp": 1612690513541,
                    "assethash": "0x70e2301955bf1e74cbb31d18c2f96972abadb328",
                    "transferaddress": "NgaiKFjurmNmiRzDRQGs44yzByXuSkdGPF",
                    "amount": "10000000000",
                    "blockindex": 3,
                    "transfernotifyindex": 0,
                    "txhash": "0xe42108b343626035cb51fbcb54949bb38aac50c8ba278841d304e9fdce0807ac"
                }
            ],
            "address": "NikhQp1aAD1YFCiwknhM5LQQebj4464bCJ"
        }

        self.mock_response(captured)

        response = await self.client.get_nep17_transfers("bogus_addr")
        self.assertEqual(2, len(response.received))
        self.assertEqual(100, response.received[0].amount)

    async def test_get_raw_mempool(self):
        tx1_hash = "0x0c65fbfd2598aee5f30cd18f1264b458f1db137c4a460f4a174facb3f2d59d06"
        tx2_hash = "0xc8040c285aa495f5b5e5b3761fd9333899f4ed902951c46d86c3bbb1cb12f2c0"

        resp = {
            "height": 5882071,
            "verified": [tx1_hash, tx2_hash],
            "unverified": []
        }
        self.mock_response(resp)

        expected_verified = [
            types.UInt256.from_string(tx1_hash[2:]),
            types.UInt256.from_string(tx2_hash[2:])
        ]
        response = await self.client.get_raw_mempool()
        self.assertEqual(expected_verified, response.verified)

    async def test_get_next_block_validators(self):
        votes = 660174
        pub_key_raw = "02237309a0633ff930d51856db01d17c829a5b2e5cc2638e9c03b4cfa8e9c9f971"
        public_key = cryptography.ECPoint.deserialize_from_bytes(bytes.fromhex(pub_key_raw))

        resp = [{
            "publickey": pub_key_raw,
            "votes": str(votes),
            "active": False
        }]
        self.mock_response(resp)

        response = await self.client.get_next_blockvalidators()
        self.assertEqual(1, len(response.validators))
        self.assertEqual(votes, response.validators[0].votes)
        self.assertEqual(public_key, response.validators[0].public_key)

    async def test_get_peers(self):
        peer1 = {"address": "47.90.28.99", "port": 21333}
        peer2 = {"address": "47.91.28.99", "port": 22333}

        resp = {
            "unconnected": [],
            "bad": [],
            "connected": [peer1, peer2]
        }
        self.mock_response(resp)

        response = await self.client.get_peers()
        self.assertEqual(2, len(response.connected))
        self.assertEqual(peer1['address'], response.connected[0].address)
        self.assertEqual(peer1['port'], response.connected[0].port)
        self.assertEqual(peer2['address'], response.connected[1].address)
        self.assertEqual(peer2['port'], response.connected[1].port)

    async def test_get_storage(self):
        storage_value = b'\x01\x02\x03'

        self.mock_response(base64.b64encode(storage_value).decode())

        dummy_contract_hash = types.UInt160.from_string("0x99042d380f2b754175717bb932a911bc0bb0ad7d"[2:])
        response = await self.client.get_storage(dummy_contract_hash, key=b'\x11')
        self.assertEqual(storage_value, response)

    async def test_get_transaction(self):
        tx = payloads.Transaction._serializable_init()
        # set some properties to ensure the TX will pass the deserialization checks
        tx.signers.append(payloads.Signer._serializable_init())
        tx.script = b'\x01'
        tx.witnesses = [payloads.Witness._serializable_init()]

        self.mock_response(base64.b64encode(tx.to_array()).decode())

        response_tx = await self.client.get_transaction(tx.hash())
        self.assertEqual(tx, response_tx)

    async def test_get_tx_height(self):
        height = 1
        tx_hash = "0c65fbfd2598aee5f30cd18f1264b458f1db137c4a460f4a174facb3f2d59d06"
        self.mock_response(height)
        response = await self.client.get_transaction_height(tx_hash)
        self.assertEqual(height, response)

    async def test_get_unclaimed_gas_(self):
        addr = "NgaiKFjurmNmiRzDRQGs44yzByXuSkdGPF"
        value = 100
        self.mock_response({"unclaimed": str(value)})
        response_value = await self.client.get_unclaimed_gas(addr)
        self.assertEqual(value, response_value)

    async def test_get_version(self):
        user_agent = "/Neo:3.0.3/"
        captured = {
            "tcpport": 10333,
            "wsport": 10334,
            "nonce": 1930156121,
            "useragent": user_agent,
            "protocol": {
                "addressversion": 53,
                "network": 860833102,
                "validatorscount": 7,
                "msperblock": 15000,
                "maxtraceableblocks": 2102400,
                "maxvaliduntilblockincrement": 5760,
                "maxtransactionsperblock": 512,
                "memorypoolmaxtransactions": 50000,
                "initialgasdistribution": 5200000000000000
            }
        }
        self.mock_response(captured)
        response = await self.client.get_version()
        self.assertEqual(user_agent, response.useragent)

    async def test_invoke_contract_verify(self):
        captured = {
            "script": "VgEMFFbIjRQK0swPKQN90Qp/AGCitShcYEBXAANAQZv2Z84MBWhlbGxvDAV3b3JsZFNB5j8YhEBXAQAMFFbIjRQK0swPKQN90Qp/AGCitShcQfgn7IxwaEA=",
            "state": "HALT",
            "gasconsumed": "1017810",
            "exception": None,
            "stack": [
                {
                    "type": "Boolean",
                    "value": True
                }
            ]
        }

        self.mock_response(captured)

        contract_hash = "92f5c79b88560584a900cfec15b0e00dc4d58b59"
        response = await self.client.invoke_contract_verify(contract_hash)
        self.assertEqual(1, len(response.stack))
        self.assertTrue(response.stack[0].value)

    async def test_invoke_function(self):
        captured = {
            "script": "VgEMFFbIjRQK0swPKQN90Qp/AGCitShcYEBXAANAQZv2Z84MBWhlbGxvDAV3b3JsZFNB5j8YhEBXAQAMFFbIjRQK0swPKQN90Qp/AGCitShcQfgn7IxwaEA=",
            "state": "HALT",
            "gasconsumed": "1017810",
            "exception": None,
            "stack": [
                {
                    "type": "Boolean",
                    "value": True
                },
                {
                    "type": "Map",
                    "value": [
                        {
                            "key": {
                                "type": "ByteString",
                                "value": "bmFtZQ=="
                            },
                            "value": {
                                "type": "ByteString",
                                "value": "TyAjOTU3IEludGVyb3BlcmFiaWxpdHk="
                            }
                        },
                        {
                            "key": {
                                "type": "ByteString",
                                "value": "b3duZXI="
                            },
                            "value": {
                                "type": "ByteString",
                                "value": "/d5rikUuvYmC6NRzIP3I3ETAoOw="
                            }
                        },
                        {
                            "key": {
                                "type": "ByteString",
                                "value": "bnVtYmVy"
                            },
                            "value": {
                                "type": "Integer",
                                "value": "957"
                            }
                        },
                        {
                            "key": {
                                "type": "ByteString",
                                "value": "aW1hZ2U="
                            },
                            "value": {
                                "type": "ByteString",
                                "value": "aHR0cHM6Ly9uZW8ub3JnL0ludGVyb3BlcmFiaWxpdHkucG5n"
                            }
                        },
                        {
                            "key": {
                                "type": "ByteString",
                                "value": "dmlkZW8="
                            },
                            "value": {
                                "type": "Any"
                            }
                        }
                    ]
                }
            ]
        }

        self.mock_response(captured)

        contract_hash = "92f5c79b88560584a900cfec15b0e00dc4d58b59"
        response = await self.client.invoke_function(contract_hash, "dummy_func")
        self.assertEqual(2, len(response.stack))
        self.assertTrue(response.stack[0].value)
        self.assertEqual("Map", response.stack[1].type)
        self.assertEqual(5, len(response.stack[1].value))

    async def test_invoke_script(self):
        captured = {
            "script": "VgEMFFbIjRQK0swPKQN90Qp/AGCitShcYEBXAANAQZv2Z84MBWhlbGxvDAV3b3JsZFNB5j8YhEBXAQAMFFbIjRQK0swPKQN90Qp/AGCitShcQfgn7IxwaEA=",
            "state": "HALT",
            "gasconsumed": "1017810",
            "exception": None,
            "stack": [
                {
                    "type": "Boolean",
                    "value": True
                }
            ]
        }

        self.mock_response(captured)

        bogus_script = b'\x01'
        response = await self.client.invoke_script(bogus_script)
        self.assertEqual(1, len(response.stack))
        self.assertTrue(response.stack[0].value)

    async def test_send_tx(self):
        tx = payloads.Transaction._serializable_init()
        hash_ = "0x13ccdb9f7eda95a24aa5a4841b24fed957fe7f1b944996cbc2e92a4fa4f1fa73"
        hash_type = types.UInt256.from_string(hash_[2:])
        bogus_response = {"hash": hash_}
        self.mock_response(bogus_response)
        response_tx_id = await self.client.send_transaction(tx)
        self.assertEqual(hash_type, response_tx_id)

    async def test_send_block(self):
        block = payloads.Block._serializable_init()
        hash_ = "0x13ccdb9f7eda95a24aa5a4841b24fed957fe7f1b944996cbc2e92a4fa4f1fa73"
        hash_type = types.UInt256.from_string(hash_[2:])
        bogus_response = {"hash": hash_}
        self.mock_response(bogus_response)
        response_block_id = await self.client.send_block(block)
        self.assertEqual(hash_type, response_block_id)

    async def test_validate_address(self):
        addr = "NPvKVTGZapmFWABLsyvfreuqn73jCjJtN1"
        resp = {
            "address": addr,
            "isvalid": True
        }
        self.mock_response(resp)
        response = await self.client.validate_address(addr)
        self.assertTrue(response)

    async def test_exception(self):
        with self.assertRaises(api.JsonRpcError) as context:
            error_response = {
                "jsonrpc": "2.0",
                "id": 1,
                "error": {
                    "code": -2146233086,
                    "message": "bogus_message",
                    "data": "bogus_data"
                }
            }
            self.helper.post('localhost', payload=error_response)
            await self.client.validate_address("abc")
        self.assertIn("-2146233086", str(context.exception))
        self.assertIn("bogus_message", str(context.exception))
        self.assertIn("bogus_data", str(context.exception))


from __future__ import annotations
from neo3 import contracts, storage, vm
from neo3.network import payloads
from neo3.core import types, cryptography, IInteroperable, serialization, to_script_hash
from typing import Any, Dict, cast, List, Tuple, Type, Optional, Union
import enum
from contextlib import suppress


class ApplicationEngine(vm.ApplicationEngineCpp):
    #: Amount of free GAS added to the engine.
    GAS_FREE = 0
    #: Maximum length of event names for "System.Runtime.Notify" SYSCALLs.
    MAX_EVENT_SIZE = 32
    #: Maximum messasge length for "System.Runtime.Log" SYSCALLs.
    MAX_NOTIFICATION_SIZE = 1024
    #: Maximum size of the smart contract script.
    MAX_CONTRACT_LENGTH = 1024 * 1024

    def __init__(self,
                 trigger: contracts.TriggerType,
                 container: payloads.IVerifiable,
                 snapshot: storage.Snapshot,
                 gas: int,
                 test_mode: bool = False
                 ):
        # Do not use super() version, see
        # https://pybind11.readthedocs.io/en/master/advanced/classes.html#overriding-virtual-functions-in-python
        vm.ApplicationEngineCpp.__init__(self, test_mode)
        #: A ledger snapshot to use for syscalls such as "System.Blockchain.GetHeight".
        self.snapshot = snapshot
        #: The trigger to run the engine with.
        self.trigger = trigger
        #: A flag to toggle infinite gas
        self.is_test_mode = test_mode

        self.script_container = container
        #: Gas available for consumption by the engine while executing its script.
        self.gas_amount = self.GAS_FREE + gas
        self._invocation_counter: Dict[types.UInt160, int] = {}
        #: Notifications (Notify SYSCALLs) that occured while executing the script.
        self.notifications: List[Tuple[payloads.IVerifiable, types.UInt160, bytes, vm.ArrayStackItem]] = []
        if self.snapshot is None or self.snapshot.persisting_block is None or self.snapshot.persisting_block.index == 0:
            self.exec_fee_factor = contracts.PolicyContract().DEFAULT_EXEC_FEE_FACTOR
            self.STORAGE_PRICE = contracts.PolicyContract().DEFAULT_STORAGE_PRICE
        else:
            self.exec_fee_factor = contracts.PolicyContract().get_exec_fee_factor(snapshot)
            self.STORAGE_PRICE = contracts.PolicyContract().get_storage_price(snapshot)

        if isinstance(self.script_container, payloads.Transaction):
            self.nonce_data = self.script_container.hash().to_array()[:16]
        else:
            self.nonce_data = b'\x00' * 16
        if self.snapshot is not None and self.snapshot.persisting_block is not None:
            nonce = self.snapshot.persisting_block.nonce
            nonce ^= int.from_bytes(self.nonce_data, "little", signed=False)
            self.nonce_data = nonce.to_bytes(16, "little")

        from neo3.contracts import interop
        self.interop = interop

    @property
    def current_scripthash(self) -> types.UInt160:
        """
        Get the script hash of the current executing smart contract

        Note: a smart contract can call other smart contracts.
        """
        if len(self.current_context.scripthash_bytes) == 0:
            return to_script_hash(self.current_context.script._value)
        return types.UInt160(self.current_context.scripthash_bytes)

    @property
    def calling_scripthash(self) -> types.UInt160:
        """
        Get the script hash of the smart contract that called the current executing smart contract.

        Note: a smart contract can call other smart contracts.

        Raises:
            ValueError: if the current executing contract has not been called by another contract.
        """
        if len(self.current_context.calling_scripthash_bytes) == 0:
            raise ValueError("Cannot retrieve calling script_hash - current context has not yet been called")
        return types.UInt160(self.current_context.calling_scripthash_bytes)

    @property
    def entry_scripthash(self) -> types.UInt160:
        """
        Get the script hash of the first smart contract loaded into the engine

        Note: a smart contract can call other smart contracts.
        """
        if len(self.entry_context.scripthash_bytes) == 0:
            return to_script_hash(self.entry_context.script._value)
        return types.UInt160(self.entry_context.scripthash_bytes)

    def checkwitness(self, hash_: types.UInt160) -> bool:
        """
        Check if the hash is a valid witness for the engines script_container
        """
        with suppress(ValueError):
            if hash_ == self.calling_scripthash:
                return True

        if isinstance(self.script_container, payloads.Transaction):
            tx = self.script_container

            response = tx.try_get_attribute(payloads.OracleResponse)
            if response is None:
                signers = tx.signers
            else:
                signers = []
                request = contracts.OracleContract().get_request(self.snapshot, response.id)
                if request:
                    tmp_tx = contracts.LedgerContract().get_tx_for_contract(self.snapshot, request.original_tx_id)
                    if tmp_tx:
                        signers = tmp_tx.signers

            for s in signers:
                if s.account == hash_:
                    signer = s
                    break
            else:
                return False

            if signer.scope == payloads.WitnessScope.GLOBAL:
                return True

            if payloads.WitnessScope.CALLED_BY_ENTRY in signer.scope:
                if self.calling_scripthash == self.entry_scripthash:
                    return True

            if payloads.WitnessScope.CUSTOM_CONTRACTS in signer.scope:
                if self.current_scripthash in signer.allowed_contracts:
                    return True

            if payloads.WitnessScope.CUSTOM_GROUPS in signer.scope:
                self._validate_callflags(contracts.CallFlags.READ_STATES)

                contract = contracts.ManagementContract().get_contract(self.snapshot, self.current_scripthash)
                if contract is None:
                    return False
                group_keys = set(map(lambda g: g.public_key, contract.manifest.groups))
                if any(group_keys.intersection(signer.allowed_groups)):
                    return True
            return False

        self._validate_callflags(contracts.CallFlags.READ_STATES)

        # for other IVerifiable types like Block
        hashes_for_verifying = self.script_container.get_script_hashes_for_verifying(self.snapshot)
        return hash_ in hashes_for_verifying

    def call_from_native(self,
                         calling_scripthash: types.UInt160,
                         hash_: types.UInt160,
                         method: str,
                         args: List[vm.StackItem]) -> None:
        ctx = self.current_context
        self._contract_call_internal(hash_, method, contracts.CallFlags.ALL, False, args)
        self.current_context.calling_scripthash_bytes = calling_scripthash.to_array()
        while self.current_context != ctx:
            self.step_out()

    def on_syscall(self, method_id: int) -> Any:
        """
        Handle interop syscalls.

        Args:
            method_id: unique syscall identifier.

        Raise:
            KeyError: if `method_id` is syscall that is not registered with the engine.
            ValueError: if the requested syscall handler is called with the wrong call flags.
            ValueError: if engine stack parameter to native type conversion fails

        Returns:
            The result of the syscall handler
        """
        descriptor = self.interop.InteropService.get_descriptor(method_id)
        if descriptor is None:
            raise KeyError(f"Requested interop {method_id} is not valid")

        self._validate_callflags(descriptor.required_call_flags)
        self.add_gas(descriptor.price * self.exec_fee_factor)

        parameters = []
        for target_type in descriptor.parameters:
            try:
                item = self.pop()
                parameters.append(self._stackitem_to_native(item, target_type))
            except IndexError:
                raise ValueError("Failed to pop parameter from stack")
            except Exception:
                raise ValueError(f"Failed to convert parameter stack item '{item}' to type '{target_type}'")

        if len(parameters) > 0:
            return_value = descriptor.handler(self, *parameters)
        else:
            return_value = descriptor.handler(self)
        if descriptor.has_return_value:
            self.push(self._native_to_stackitem(return_value, type(return_value)))
        return return_value

    def invoke_syscall_by_name(self, method: str) -> Any:
        """
        Helper function to call `on_syscall` using the syscall name.

        Args:
            method: full qualified syscall name. e.g. "System.Runtime.Platform"

        Returns: the result of the syscall handler. e.g. for "System.Runtime.Platform" returns "NEO"
        """
        return self.on_syscall(contracts.syscall_name_to_int(method))

    def get_invocation_counter(self) -> int:
        """
        Get the number of times the current contract has been called during this execute() run.

        Note: the counter increases with every "System.Contract.Call" SYSCALL

        Raises:
            ValueError: if the contract has not been called.
        """
        counter = self._invocation_counter.get(self.current_scripthash, None)
        if counter is None:
            self._invocation_counter.update({self.current_scripthash: 1})
            counter = 1
        return counter

    def load_context(self, context: vm.ExecutionContext) -> None:
        if len(context.scripthash_bytes) == 0:
            context.scripthash_bytes = to_script_hash(context.script._value).to_array()
        contract_hash = types.UInt160(data=context.scripthash_bytes)
        counter = self._invocation_counter.get(contract_hash, 0)
        self._invocation_counter.update({contract_hash: counter + 1})

        super(ApplicationEngine, self).load_context(context)

    def load_contract(self,
                      contract: contracts.ContractState,
                      method_descriptor: contracts.ContractMethodDescriptor,
                      flags: contracts.CallFlags) -> Optional[vm.ExecutionContext]:

        rvcount = 0 if method_descriptor.return_type == contracts.ContractParameterType.VOID else 1
        context = self.load_script_with_callflags(vm.Script(contract.script),
                                                  flags,
                                                  method_descriptor.offset,
                                                  rvcount)
        # configure state
        context.call_flags = int(flags)
        context.scripthash_bytes = contract.hash.to_array()
        context.nef_bytes = contract.nef.to_array()

        init = contract.manifest.abi.get_method("_initialize", 0)
        if init is not None:
            self.load_context(context.clone(init.offset))
        return context

    def load_token(self, token_id: int) -> vm.ExecutionContext:
        self._validate_callflags(contracts.CallFlags.READ_STATES | contracts.CallFlags.ALLOW_CALL)
        if len(self.current_context.nef_bytes) == 0:
            raise ValueError("Current context has no NEF state")
        nef = contracts.NEF.deserialize_from_bytes(self.current_context.nef_bytes)
        if token_id >= len(nef.tokens):
            raise ValueError("token_id exceeds available tokens")

        token = nef.tokens[token_id]
        if token.parameters_count > len(self.current_context.evaluation_stack):
            raise ValueError("Token count exceeds available parameters on evaluation stack")
        args: List[vm.StackItem] = []
        for _ in range(token.parameters_count):
            args.append(self.pop())
        return self._contract_call_internal(token.hash, token.method, token.call_flags, token.has_return_value, args)

    def load_script_with_callflags(self,
                                   script: vm.Script,
                                   call_flags: contracts.CallFlags,
                                   initial_position: int = 0,
                                   rvcount: int = -1):
        context = super(ApplicationEngine, self).load_script(script, rvcount, initial_position)
        context.call_flags = int(call_flags)
        return context

    def step_out(self) -> None:
        c = len(self.invocation_stack)
        while self.state != vm.VMState.HALT and self.state != vm.VMState.FAULT and len(self.invocation_stack) >= c:
            self._execute_next()
        if self.state == vm.VMState.FAULT:
            raise ValueError(f"Call from native contract failed: {self.exception_message}")

    def _contract_call_internal(self,
                                contract_hash: types.UInt160,
                                method: str,
                                flags: contracts.CallFlags,
                                has_return_value: bool,
                                args: List[vm.StackItem]) -> vm.ExecutionContext:
        target_contract = contracts.ManagementContract().get_contract(self.snapshot, contract_hash)
        if target_contract is None:
            raise ValueError("[System.Contract.Call] Can't find target contract")

        method_descriptor = target_contract.manifest.abi.get_method(method, len(args))
        if method_descriptor is None:
            raise ValueError(f"[System.Contract.Call] Method '{method}' with {len(args)} arguments does not exist on "
                             f"target contract")
        return self._contract_call_internal2(target_contract, method_descriptor, flags, has_return_value, args)

    def _contract_call_internal2(self,
                                 target_contract: contracts.ContractState,
                                 method_descriptor: contracts.ContractMethodDescriptor,
                                 flags: contracts.CallFlags,
                                 has_return_value: bool,
                                 args: List[vm.StackItem]):
        if contracts.PolicyContract().is_blocked(self.snapshot, target_contract.hash):
            raise ValueError(f"[System.Contract.Call] Can't call blocked contract")
        if method_descriptor.safe:
            flags &= ~(contracts.CallFlags.WRITE_STATES | contracts.CallFlags.ALLOW_NOTIFY)
        else:
            current_contract = contracts.ManagementContract().get_contract(self.snapshot, self.current_scripthash)
            if current_contract and not current_contract.can_call(target_contract, method_descriptor.name):
                raise ValueError(
                    f"[System.Contract.Call] Not allowed to call target method '{method_descriptor.name}' according "
                    f"to manifest")

        counter = self._invocation_counter.get(target_contract.hash, 0)
        self._invocation_counter.update({target_contract.hash: counter + 1})

        state = self.current_context
        calling_script_hash_bytes = state.scripthash_bytes
        calling_flags = state.call_flags

        arg_len = len(args)
        expected_len = len(method_descriptor.parameters)
        if arg_len != expected_len:
            raise ValueError(
                f"[System.Contract.Call] Invalid number of contract arguments. Expected {expected_len} actual {arg_len}")  # noqa

        if has_return_value ^ (method_descriptor.return_type != contracts.ContractParameterType.VOID):
            raise ValueError("Return value type does not match")

        context_new = self.load_contract(target_contract,
                                         method_descriptor,
                                         flags & calling_flags)
        if context_new is None:
            raise ValueError
        context_new.calling_scripthash_bytes = calling_script_hash_bytes

        for item in reversed(args):
            context_new.evaluation_stack.push(item)

        return context_new

    def _stackitem_to_native(self, stack_item: vm.StackItem, target_type: Type[object]):
        # checks for type annotations like `List[bytes]` (similar to byte[][] in C#)
        if hasattr(target_type, '__origin__'):
            if target_type.__origin__ == list:  # type: ignore
                element_type = target_type.__args__[0]  # type: ignore
                array = []
                if isinstance(stack_item, vm.ArrayStackItem):
                    for e in stack_item:
                        array.append(self._convert(e, element_type))
                else:
                    count = stack_item.to_biginteger()
                    if count > self.MAX_STACK_SIZE:
                        raise ValueError

                    # mypy bug: https://github.com/python/mypy/issues/9755
                    for e in range(count):  # type: ignore
                        array.append(self._convert(self.pop(), element_type))
                return array
            if target_type.__origin__ == Union:  # type: ignore
                # handle typing.Optional[type], Optional is an alias for Union[x, None]
                # only support specifying 1 type
                if len(target_type.__args__) != 2:  # type: ignore
                    raise ValueError(f"Don't know how to convert {target_type}")
                if isinstance(stack_item, vm.NullStackItem):
                    return None
                else:
                    for i in target_type.__args__:  # type: ignore
                        if i is None:
                            continue
                        return self._convert(stack_item, i)
        else:
            try:
                return self._convert(stack_item, target_type)
            except ValueError:
                if isinstance(stack_item, vm.InteropStackItem):
                    return stack_item.get_object()
                else:
                    raise

    def _validate_callflags(self, callflags: contracts.CallFlags) -> None:
        if callflags not in contracts.CallFlags(self.current_context.call_flags):
            raise ValueError(f"Context requires callflags {callflags}")

    def _convert(self, stack_item: vm.StackItem, class_type: Type[object]) -> object:
        """
        convert VM type to native
        """
        if class_type in [vm.StackItem, vm.PointerStackItem, vm.ArrayStackItem, vm.InteropStackItem]:
            return stack_item
        elif class_type == int:
            return int(stack_item.to_biginteger())
        elif class_type == vm.BigInteger:
            return stack_item.to_biginteger()
        # mypy bug? https://github.com/python/mypy/issues/9756
        elif class_type in [bytes, bytearray]:  # type: ignore
            return stack_item.to_array()
        elif class_type == bool:
            return stack_item.to_boolean()
        elif class_type == types.UInt160:
            return types.UInt160(data=stack_item.to_array())
        elif class_type == types.UInt256:
            return types.UInt256(data=stack_item.to_array())
        elif class_type == str:
            if stack_item == vm.NullStackItem():
                return ""
            return stack_item.to_array().decode()
        elif class_type == cryptography.ECPoint:
            return cryptography.ECPoint.deserialize_from_bytes(stack_item.to_array())
        elif issubclass(class_type, enum.Enum):
            if stack_item.get_type() == vm.StackItemType.INTEGER:
                stack_item = cast(vm.IntegerStackItem, stack_item)
                # mypy seems to have trouble understanding types that support __int__
                return class_type(int(stack_item))  # type: ignore
            elif stack_item.get_type() == vm.StackItemType.BYTESTRING:
                stack_item = cast(vm.ByteStringStackItem, stack_item)
                return class_type(int(stack_item.to_biginteger()))  # type: ignore
        raise ValueError(f"Unknown class type, don't know how to convert: {class_type}")

    def _native_to_stackitem(self, value, native_type) -> vm.StackItem:
        """
        Convert native type to VM type

        Note: order of checking matters.
        e.g. a Transaction should be treated as IInteroperable, while its also ISerializable
        """
        if isinstance(value, vm.StackItem):
            return value
        elif value is None:
            return vm.NullStackItem()
        elif native_type in [int, vm.BigInteger]:
            return vm.IntegerStackItem(value)
        elif issubclass(native_type, IInteroperable):
            value_ = cast(IInteroperable, value)
            return value_.to_stack_item(self.reference_counter)
        elif issubclass(native_type, serialization.ISerializable):
            serializable_value = cast(serialization.ISerializable, value)
            return vm.ByteStringStackItem(serializable_value.to_array())
        # mypy bug? https://github.com/python/mypy/issues/9756
        elif native_type in [bytes, bytearray]:  # type: ignore
            return vm.ByteStringStackItem(value)
        elif native_type == str:
            return vm.ByteStringStackItem(bytes(value, 'utf-8'))
        elif native_type == bool:
            return vm.BooleanStackItem(value)
        elif issubclass(native_type, (enum.IntFlag, enum.IntEnum)):
            return self._native_to_stackitem(value.value, int)
        elif hasattr(native_type, '__origin__') and native_type.__origin__ == Union:  # type: ignore
            # handle typing.Optional[type], Optional is an alias for Union[x, None]
            # only support specifying 1 type
            if len(native_type.__args__) != 2:
                raise ValueError(f"Don't know how to convert native type {native_type} to stackitem")
            for i in native_type.__args__:
                if i is None:
                    continue
                return self._native_to_stackitem(value, native_type)
            else:
                raise ValueError  # shouldn't be possible, but silences mypy
        elif native_type == list:
            arr = vm.ArrayStackItem(self.reference_counter)
            for item in value:
                arr.append(self._native_to_stackitem(item, type(item)))
            return arr
        elif native_type == tuple:
            _struct = vm.StructStackItem(self.reference_counter)
            _struct.append(self._native_to_stackitem(value[0], type(value[0])))
            _struct.append(self._native_to_stackitem(value[1], type(value[1])))
            return _struct
        else:
            return vm.StackItem.from_interface(value)



import unittest
import binascii
from neo3 import contracts, vm
from neo3.core import cryptography, types


class ContractTestCase(unittest.TestCase):
    def shortDescription(self):
        # disable docstring printing in test runner
        return None

    def test_create_signature_contract(self):
        """
        var priv_key1 = new byte[]
        {
            0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1,
            0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1,
            0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1,
            0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1,
        };
        var kp1 = new KeyPair(priv_key1);
        var c = Contract.CreateSignatureContract(kp1.PublicKey);
        Console.WriteLine(c.Script.ToHexString());
        """
        expected = binascii.unhexlify(b'0c21026ff03b949241ce1dadd43519e6960e0a85b41a69a05c328103aa2bce1594ca164156e7b327')
        keypair = cryptography.KeyPair(private_key=b'\x01' * 32)
        contract = contracts.Contract.create_signature_contract(keypair.public_key)
        self.assertEqual(expected, contract.script)

    def test_create_multisignature_contract(self):
        """
        var priv_key1 = new byte[]
        {
            0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1,
            0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1,
            0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1,
            0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1,
        };
        var priv_key2 = new byte[]
        {
            0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2,
            0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2,
            0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2,
            0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2, 0x2,
        };
        var kp1 = new KeyPair(priv_key1);
        var kp2 = new KeyPair(priv_key2);
        var c = Contract.CreateMultiSigContract(1, new ECPoint[] {kp1.PublicKey, kp2.PublicKey});
        Console.WriteLine(c.Script.ToHexString());
        Console.WriteLine(c.ScriptHash.ToArray().ToHexString());
        """
        expected_script = binascii.unhexlify(b'110c2102550f471003f3df97c3df506ac797f6721fb1a1fb7b8f6f83d224498a65c88e240c21026ff03b949241ce1dadd43519e6960e0a85b41a69a05c328103aa2bce1594ca1612419ed0dc3a')
        expected_script_hash = types.UInt160(binascii.unhexlify(b'a72d98260b3dd7d7dd4f733490b9c16b5c352644'))
        keypair1 = cryptography.KeyPair(private_key=b'\x01' * 32)
        keypair2 = cryptography.KeyPair(private_key=b'\x02' * 32)
        contract = contracts.Contract.create_multisig_contract(1, [keypair1.public_key, keypair2.public_key])
        self.assertEqual(expected_script, contract.script)
        self.assertEqual(str(expected_script_hash), str(contract.script_hash))

    def test_create_multisignature_redeemscript_invalid_arguments(self):
        with self.assertRaises(ValueError) as context:
            contracts.Contract.create_multisig_redeemscript(0, [])
        self.assertEqual("Minimum required signature count is 1, specified 0.", str(context.exception))

        with self.assertRaises(ValueError) as context:
            contracts.Contract.create_multisig_redeemscript(1, [])
        self.assertEqual("Invalid public key count. Minimum required signatures is bigger than supplied public keys count.", str(context.exception))

        with self.assertRaises(ValueError) as context:
            contracts.Contract.create_multisig_redeemscript(1, [object() for _ in range(0, 1025)])
        self.assertEqual("Supplied public key count (1025) exceeds maximum of 1024.", str(context.exception))

    def test_is_signature_contract(self):
        """
        A valid signature contract script looks as follows
        - PUSHDATA1 (0xC)
        - LEN PUBLIC KEY (33)
        - PUBLIC KEY data
        - SYSCALL (0x41)
        - "System.Crypto.CheckSig" identifier
        """

        incorrect_script_len = b'\x01' * 10
        self.assertFalse(contracts.Contract.is_signature_contract(incorrect_script_len))

        # first byte should be PUSHDATA1 (0xC)
        incorrect_script_start_byte = b'\x01' * 40
        self.assertFalse(contracts.Contract.is_signature_contract(incorrect_script_start_byte))

        # second byte should be 33
        incorrect_second_byte = bytearray(b'\x01' * 40)
        incorrect_second_byte[0] = int(vm.OpCode.PUSHDATA1)
        self.assertFalse(contracts.Contract.is_signature_contract(incorrect_second_byte))

        # index 35 should be SYSCALL
        incorrect_idx_35 = bytearray([0xc, 33]) + b'\01' * 38
        incorrect_idx_35[35] = int(vm.OpCode.PUSHNULL)
        self.assertFalse(contracts.Contract.is_signature_contract(incorrect_idx_35))  # index 35 should be SYSCALL

        # the last 4 bytes should be the "System.Crypto.CheckSig" SYSCALL
        incorrect_syscall_number = bytearray([0xc, 33]) + b'\01' * 38
        incorrect_syscall_number[35] = int(vm.OpCode.SYSCALL)
        self.assertFalse(contracts.Contract.is_signature_contract(incorrect_syscall_number))

        # and finally a contract that matches the correct format
        correct = bytearray([0xc, 33]) + b'\01' * 38
        correct[35] = int(vm.OpCode.SYSCALL)
        correct[36:40] = contracts.syscall_name_to_int("System.Crypto.CheckSig").to_bytes(4, 'little')
        self.assertTrue(contracts.Contract.is_signature_contract(correct))

    def test_is_multisig_contract_too_short(self):
        script_too_short = b'\x00'
        self.assertFalse(contracts.Contract.is_multisig_contract(script_too_short))

    def test_is_multisig_contract_invalid_public_key(self):
        """
        A valid multisignature contract looks as follows
        - signature length (variable int)
        - number of public keys times
            - PUSHDATA1
            - public key len (0x33)
            - public key data
        """
        script = bytearray([int(vm.OpCode.PUSHINT8)])
        # signature count
        script += b'\x02'
        # public key 1
        script += bytearray([int(vm.OpCode.PUSHDATA1)])
        script += bytearray([33])
        script += bytes.fromhex("03a60c1deaf147b10691c344c76e5f3dac83b555fdd5a3f8d9e2f623b3d1af8df6")
        # public key 2, but the key data is too short
        script += bytearray([int(vm.OpCode.PUSHDATA1)])
        script += b'\xFF' * 10
        self.assertFalse(contracts.Contract.is_multisig_contract(script))

    def test_is_multisig_contract_invalid_public_key_2(self):
        script = bytearray([int(vm.OpCode.PUSHINT8)])
        # signature count
        script += b'\x02'
        # public key 1
        script += bytearray([int(vm.OpCode.PUSHDATA1)])
        script += bytearray([33])
        script += bytes.fromhex("03a60c1deaf147b10691c344c76e5f3dac83b555fdd5a3f8d9e2f623b3d1af8df6")
        # public key 2, but the key data is too short
        script += bytearray([int(vm.OpCode.PUSHDATA1)])
        # public key 2 length should be 33, but we make it 00
        script += b'\x00'
        script += bytes.fromhex("03a60c1deaf147b10691c344c76e5f3dac83b555fdd5a3f8d9e2f623b3d1af8df6")

        script += bytearray([int(vm.OpCode.PUSHINT8)])
        self.assertFalse(contracts.Contract.is_multisig_contract(script))
        return script

    def test_is_multisig_contract_public_key_count_lower_than_signature_count(self):
        script = bytearray([int(vm.OpCode.PUSHINT8)])
        # signature count
        script += b'\x02'
        # make sure it meets the minimum length
        script += b'\x00' * 41
        # assert that 0 public keys does not meet the signature count of 2
        self.assertFalse(contracts.Contract.is_multisig_contract(script))

    def test_public_key_count_mismatch(self):
        # the specified public key count should match the number of public keys found
        # we can re-use the script from a previous test, where we only have to fix the length field
        # for public key 2
        script = self.test_is_multisig_contract_invalid_public_key_2()
        script[-35] = 33
        self.assertFalse(contracts.Contract.is_multisig_contract(script))

    def test_script_invalid_tail(self):
        # same as previous test, we re-use and fix the length
        script = self.test_is_multisig_contract_invalid_public_key_2()
        script[-35] = 33
        # now we make sure the public key count matches
        script[-1] = int(vm.OpCode.PUSHINT8)
        script += b'\x02'
        # and assert we don't have enough data left for the remainder of the checks
        self.assertFalse(contracts.Contract.is_multisig_contract(script))

        # now we extend with 5 bytes to give enough data
        script += b'\x00' * 5
        # the first byte should be a SYSCALL, but it isn't
        self.assertFalse(contracts.Contract.is_multisig_contract(script))

        # we fix the SYSCALL byte
        script[-5] = int(vm.OpCode.SYSCALL)
        # finally test the last 4 bytes should be "System.Crypto.CheckSig" syscall number
        self.assertFalse(contracts.Contract.is_multisig_contract(script))

    def test_is_multsig_contract_ok(self):
        keypair = cryptography.KeyPair(private_key=b'\x01' * 32)
        contract = contracts.Contract.create_multisig_contract(1, [keypair.public_key])
        self.assertTrue(contracts.Contract.is_multisig_contract(contract.script))

    def test_is_multisig_contract_256_pubkeys(self):
        # test handling of a large number of public keys
        # from 256 signatures and above he count is encoded in 2 bytes
        # for this test we manually encoded a length of 2 into 2 bytes
        script = bytearray([int(vm.OpCode.PUSHINT16), 2, 0])

        # add fake public keys
        for _ in range(0, 2):
            script += bytearray([int(vm.OpCode.PUSHDATA1)])
            script += bytearray([33])
            script += bytes.fromhex("03a60c1deaf147b10691c344c76e5f3dac83b555fdd5a3f8d9e2f623b3d1af8df6")

        # and now mismatch the public key count value we say is present (0 here)
        script += bytearray([int(vm.OpCode.PUSHINT16), 0, 0])
        self.assertFalse(contracts.Contract.is_multisig_contract(script))

        # now we correct the public key count in the script and make it valid by adding the expected tail
        script[-2] = 2
        script += bytearray([int(vm.OpCode.SYSCALL)])
        script += contracts.syscall_name_to_int("System.Crypto.CheckMultisig").to_bytes(4, 'little')
        self.assertTrue(contracts.Contract.is_multisig_contract(script))

    def test_is_multsig_contract_invalid_pubkey_count(self):
        # this is another test where we check for a mismatch between processed public keys and the claimed present
        # amount but this time for a signature count between push0 <= cnt <= PUSH16
        # we re-use the standard create_multisignature_account call
        keypair = cryptography.KeyPair(private_key=b'\x01' * 32)
        contract = contracts.Contract.create_multisig_contract(1, [keypair.public_key])
        # and modify the claimed public key length from 1 to 255
        data = bytearray(contract.script)
        data[-6] = int(vm.OpCode.PUSH16) - 1
        self.assertFalse(contracts.Contract.is_multisig_contract(data))

        # and finally we try it again but this time the public key length is in an invalid range
        data[-6] = int(vm.OpCode.PUSH16) + 1
        self.assertFalse(contracts.Contract.is_multisig_contract(data))

    def test_is_multisig_contract_invalid_sig_counts(self):
        # no valid signature length byte
        script = b'\xFF' * 43
        self.assertFalse(contracts.Contract.is_multisig_contract(script))

        # more than 1024 signatures
        sig_cnt = 1025
        script = b'\x01' + sig_cnt.to_bytes(2, 'little') + b'\x01' * 40
        self.assertFalse(contracts.Contract.is_multisig_contract(script))



from __future__ import annotations
import hashlib
import enum
from . import NativeContract, register
from neo3 import contracts
from neo3.core import cryptography


class NamedCurve(enum.IntEnum):
    SECP256K1 = 22
    SECP256R1 = 23


class CryptoContract(NativeContract):

    _service_name = "CryptoLib"
    _id = -3

    curves = {
        NamedCurve.SECP256K1: cryptography.ECCCurve.SECP256K1,
        NamedCurve.SECP256R1: cryptography.ECCCurve.SECP256R1
    }

    def init(self):
        super(CryptoContract, self).init()

    @register("ripemd160", contracts.CallFlags.NONE, cpu_price=1 << 15)
    def ripemd160(self, data: bytes) -> bytes:
        return hashlib.new('ripemd160', data).digest()

    @register("sha256", contracts.CallFlags.NONE, cpu_price=1 << 15)
    def sha256(self, data: bytes) -> bytes:
        return hashlib.sha256(data).digest()

    @register("verifyWithECDsa", contracts.CallFlags.NONE, cpu_price=1 << 15)
    def verify_with_ecdsa(self, message: bytes, public_key: bytes, signature: bytes, curve: NamedCurve) -> bool:
        return cryptography.verify_signature(message, signature, public_key, self.curves.get(curve))



import unittest
from neo3 import vm, contracts, storage
from neo3.network import payloads
from neo3.core import to_script_hash, types, cryptography
from tests.contracts.interop.utils import test_engine, test_tx
from copy import deepcopy

"""
We compile the following smart contract for testing using neo3-boa

@public
def main() -> str:
   return "hello world"
"""

raw_hello_world_nef = b'NEF3neo3-boa by COZ-0.7.0.0\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x0e\x0c\x0bhello world@_h.\xd7'
raw_hello_world_manifest = {
    "name": "hello_world",
    "groups": [],
    "features": {},
    "abi": {
        "methods": [
            {
                "name": "main",
                "offset": 0,
                "parameters": [],
                "returntype": "String",
                "safe": False
            }
        ],
        "events": []
    },
    "permissions": [
        {
            "contract": "*",
            "methods": "*"
        }
    ],
    "trusts": [],
    "supportedstandards": [],
    "extra": None
}

hello_world_nef = contracts.NEF.deserialize_from_bytes(raw_hello_world_nef)
hello_world_manifest = contracts.manifest.ContractManifest.from_json(raw_hello_world_manifest)

raw_bye_world_nef = b'NEF3neo3-boa by COZ-0.7.0.0\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x0c\x0c\tbye world@NLE\xa8'
raw_bye_world_manifest = {
    "name": "bye_world",
    "groups": [],
    "features": {},
    "abi": {
        "methods": [
            {
                "name": "main",
                "offset": 0,
                "parameters": [],
                "returntype": "String",
                "safe": False
            }
        ],
        "events": []
    },
    "permissions": [
        {
            "contract": "*",
            "methods": "*"
        }
    ],
    "trusts": [],
    "supportedstandards": [],
    "extra": None
}
bye_world_nef = contracts.NEF.deserialize_from_bytes(raw_bye_world_nef)
bye_world_manifest = contracts.ContractManifest.from_json(raw_bye_world_manifest)


"""
Contract3 code - compiled with neo3-boa

from boa3.builtin import public

def main() -> int:
    return 1

@public
def test_func() -> int:
    return 2

@public
def test_func2(value: int) -> int:
    return 1 + value
"""
raw_contract3_nef = b'NEF3neo3-boa by COZ-0.7.0.0\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x0b\x11@\x12@W\x00\x01\x11x\x9e@\xaa,\x0bW'
raw_contract3_manifest = {
    "name": "contract3",
    "groups": [],
    "features": {},
    "abi": {
        "methods": [
            {
                "name": "test_func",
                "offset": 2,
                "parameters": [],
                "returntype": "Integer",
                "safe": False
            },
            {
                "name": "test_func2",
                "offset": 4,
                "parameters": [
                    {
                        "name": "value",
                        "type": "Integer"
                    }
                ],
                "returntype": "Integer",
                "safe": False
            }
        ],
        "events": []
    },
    "permissions": [
        {
            "contract": "*",
            "methods": "*"
        }
    ],
    "trusts": [],
    "supportedstandards": [],
    "extra": None
}
contract3_nef = contracts.NEF.deserialize_from_bytes(raw_contract3_nef)
contract3_manifest = contracts.ContractManifest.from_json(raw_contract3_manifest)


class RuntimeInteropTestCase(unittest.TestCase):
    def shortDescription(self):
        # disable docstring printing in test runner
        return None

    def test_validate_sample_contract(self):
        engine = test_engine()
        engine.load_script(vm.Script(hello_world_nef.script))
        engine.execute()
        self.assertEqual(vm.VMState.HALT, engine.state)
        self.assertEqual(1, len(engine.result_stack._items))
        item = engine.result_stack.pop()
        self.assertIsInstance(item, vm.ByteStringStackItem)
        self.assertEqual(vm.ByteStringStackItem(b'hello world'), item)

    @unittest.SkipTest
    def test_contract_create_invalid_manifest_or_script(self):
        # script len 0
        engine = test_engine(has_snapshot=True)
        engine.script_container = test_tx(1)
        invalid_script = b''
        engine.push(vm.ByteStringStackItem(str(hello_world_manifest).encode()))
        engine.push(vm.ByteStringStackItem(invalid_script))
        with self.assertRaises(ValueError) as context:
            engine.invoke_syscall_by_name("System.Contract.Create")
        self.assertEqual("Invalid script or manifest length", str(context.exception))

        # script too long
        invalid_script = b'\x00' * (engine.MAX_CONTRACT_LENGTH + 1)
        engine.push(vm.ByteStringStackItem(str(hello_world_manifest).encode()))
        engine.push(vm.ByteStringStackItem(invalid_script))
        with self.assertRaises(ValueError) as context:
            engine.invoke_syscall_by_name("System.Contract.Create")
        self.assertEqual("Invalid script or manifest length", str(context.exception))

        # manifest len 0
        invalid_manifest = b''
        engine.push(vm.ByteStringStackItem(invalid_manifest))
        engine.push(vm.ByteStringStackItem(hello_world_nef.script))
        with self.assertRaises(ValueError) as context:
            engine.invoke_syscall_by_name("System.Contract.Create")
        self.assertEqual("Invalid script or manifest length", str(context.exception))

        # manifest too long
        invalid_manifest = b'\x00' * (contracts.ContractManifest.MAX_LENGTH + 1)
        engine.push(vm.ByteStringStackItem(invalid_manifest))
        engine.push(vm.ByteStringStackItem(hello_world_nef.script))
        with self.assertRaises(ValueError) as context:
            engine.invoke_syscall_by_name("System.Contract.Create")
        self.assertEqual("Invalid script or manifest length", str(context.exception))

    def test_contract_call(self):
        engine = test_engine(has_snapshot=True, default_script=False)
        # current executing contract
        fake_hash = types.UInt160.deserialize_from_bytes(b'\x01' * 20)
        contract = contracts.ContractState(0, hello_world_nef, hello_world_manifest, 0, fake_hash)
        engine.snapshot.contracts.put(contract)
        # target contract
        fake_hash2 = types.UInt160.deserialize_from_bytes(b'\x02' * 20)
        target_contract = contracts.ContractState(1, contract3_nef, contract3_manifest, 0, fake_hash2)
        engine.snapshot.contracts.put(target_contract)
        engine.load_script(vm.Script(contract.script))
        array = vm.ArrayStackItem(engine.reference_counter)
        array.append(vm.IntegerStackItem(3))
        engine.push(array)  # args
        engine.push(vm.IntegerStackItem(15))  # callflags
        engine.push(vm.ByteStringStackItem("test_func2"))  # method
        engine.push(vm.ByteStringStackItem(target_contract.hash.to_array()))
        engine.invoke_syscall_by_name("System.Contract.Call")
        engine.execute()

        self.assertEqual(2, len(engine.result_stack))
        main_contract_return_value = engine.result_stack.pop()
        syscall_called_contract_return_value = engine.result_stack.pop()
        self.assertEqual("hello world", main_contract_return_value.to_array().decode())
        self.assertEqual(4, int(syscall_called_contract_return_value))

    def test_contract_call_exceptions(self):
        engine = test_engine(has_snapshot=True, default_script=False)
        engine.load_script(vm.Script(hello_world_nef.script))

        # can't find contract
        with self.assertRaises(ValueError) as context:
            engine._contract_call_internal(types.UInt160.zero(), "valid_method", contracts.CallFlags.ALL, False, vm.ArrayStackItem(engine.reference_counter))
        self.assertEqual("[System.Contract.Call] Can't find target contract", str(context.exception))

        fake_contract_hash = types.UInt160(b'\x01' * 20)
        target_contract = contracts.ContractState(0, contract3_nef, contract3_manifest, 0, fake_contract_hash)
        engine.snapshot.contracts.put(target_contract)

        # modify the manifest of the current executing contract to only allow to call 1 specific method on other contracts
        new_current_manifest = deepcopy(hello_world_manifest)
        new_current_manifest.permissions = [contracts.ContractPermission(
            contracts.ContractPermissionDescriptor(),  # allow to call any contract
            contracts.WildcardContainer(['method_aaaa'])  # allowing to call the listed method only
        )]
        fake_contract_hash2 = types.UInt160(b'\x02' * 20)
        new_current_contract = contracts.ContractState(1, hello_world_nef, new_current_manifest, 0, fake_contract_hash2)
        engine.snapshot.contracts.put(new_current_contract)
        with self.assertRaises(ValueError) as context:
            engine._contract_call_internal(target_contract.hash, "invalid_method", contracts.CallFlags.ALL, False, vm.ArrayStackItem(engine.reference_counter))
        self.assertEqual("[System.Contract.Call] Method 'invalid_method' with 0 arguments does not exist on target contract", str(context.exception))

        # restore current contract to its original form and try to call a non-existing contract
        current_contract = contracts.ContractState(1, hello_world_nef, hello_world_manifest, 1, fake_contract_hash2)
        engine.snapshot.contracts.delete(new_current_contract.hash)
        engine.snapshot.contracts.put(current_contract)

        with self.assertRaises(ValueError) as context:
            engine._contract_call_internal(target_contract.hash, "invalid_method", contracts.CallFlags.ALL, False, vm.ArrayStackItem(engine.reference_counter))
        self.assertEqual("[System.Contract.Call] Method 'invalid_method' with 0 arguments does not exist on target contract", str(context.exception))

        # call the target method with invalid number of arguments
        array = vm.ArrayStackItem(engine.reference_counter)
        array.append([vm.NullStackItem(), vm.NullStackItem()])
        with self.assertRaises(ValueError) as context:
            engine._contract_call_internal(target_contract.hash, "test_func", contracts.CallFlags.ALL, False, array)
        self.assertEqual("[System.Contract.Call] Method 'test_func' with 2 arguments does not exist on target contract", str(context.exception))

    def test_contract_call_flags(self):
        engine = test_engine()
        engine.invoke_syscall_by_name("System.Contract.GetCallFlags")
        self.assertEqual(1, len(engine.current_context.evaluation_stack))
        self.assertEqual(contracts.CallFlags.ALL, contracts.CallFlags(int(engine.pop())))

    def test_contract_create_standard_account(self):
        keypair = cryptography.KeyPair(b'\x01' * 32)
        engine = test_engine()
        engine.push(vm.ByteStringStackItem(keypair.public_key.to_array()))
        engine.invoke_syscall_by_name("System.Contract.CreateStandardAccount")
        engine.execute()
        self.assertEqual(1, len(engine.result_stack))
        signature_redeem_script = contracts.Contract.create_signature_redeemscript(keypair.public_key)
        result_item = types.UInt160(engine.result_stack.pop().to_array())
        self.assertEqual(to_script_hash(signature_redeem_script), result_item)



import unittest

from neo3.wallet import Account

account_list = [
    {
        "address": "NRaKbRA5JAEJtfUgJJZzmeDnKvP3pJwKp1",
        "encrypted_key": "6PYKuriAL7pFeVTr3tKksbD1SpKUP7K82vjGuskZ5zpo9EWDhLRW6GcnyL",
        "password": "city of zion",
        "private_key": "58124574dfcca1a7a958775f6ea94e3d6c392ec3ba125b5bc591dd5e14f05e52",
        "script_hash": "18f13748e08d53c9a164227e1a3e8d8d9e78193e",
        "wif_key": "KzAuju4yBqBhmUzYpfEEppPW8jfxALTsdsUR8hLPv9R3PBD97CUv"      
    },
    {
        "address": "NgPptMp2tcjnXuYbUrTozvwvLExGKk5jXc",
        "encrypted_key": "6PYMEujkLs249znmr3v59x3M12iPbdPTZftv8f2TapuLyHn8TQBZLVJUV6",
        "password": "123",
        "private_key": "2032b737522d22e2b6faf30555faa91d95c5aa5113c18f218f45815b6934c558",
        "script_hash": "cfa9032d65b3d0fc1df3956a4ef01666f23ba7e0",
        "wif_key": "KxJJLmU1Nv7igx3RFM4siSvio7wasF3ZzMzi7SrJ1s78QDQeEtjs"
    },
    {
        "address": "NZMHRJMPbyJJwtXpvS2mYAWcWp4qmZZFx8",
        "encrypted_key": "6PYL44vbS5e6ubYtV3JqDM7J92gCEYXWewVyFdyki6JLcQ7QaYzsW6YjTs",
        "password": "neo",
        "private_key": "4c5182d9041f416bee1a6adac6a03f3e0319a83e75e78e6ff739304095791f19",
        "script_hash": "0df27baba6baeeb6834bea0d6c2a78183b416393",
        "wif_key": "Kyn4fA6czAhktoAM9YXKv3m7jtt47AuQxCXqSusnBmj3GsZUZQ6M"
    },
    {
        "address": "NWxsLx9BFA558pVLZmFNuYsRKuXMMi2QSQ",
        "encrypted_key": "6PYQ5gTPAv4p5nfrb2ywPumSdtiXCjimuwDAbv93FPancd1dU9D9ajkPqd",
        "password": "neo-mamba",
        "private_key": "1c43f87ce2ce3ea676bdfec4928705f3e9fedbdb3acf1fed6b3cc0c3d87c4cad",
        "script_hash": "3b951421e8dc81552df3af1478ef72b05bc13579",
        "wif_key": "KxAexpL1F8FAoG72LCDkerLa8SKLutUSLafALB5zvdQ2asuRf6Wx"
    }
]


class AccountCreationTestCase(unittest.TestCase):

    def test_new_account(self):
        for testcase in account_list:
            account = Account(testcase['password'])
            self.assertIsNotNone(account)
            self.assertIsNotNone(account.address)
            self.assertIsNotNone(account.encrypted_key)
            self.assertIsNotNone(account.public_key)

    def test_new_account_from_private_key(self):
        for testcase in account_list:
            account = Account.from_private_key(bytes.fromhex(testcase['private_key']), testcase['password'])
            self.assertEqual(testcase['address'], account.address)
            self.assertEqual(testcase['encrypted_key'].encode('utf-8'), account.encrypted_key)
            self.assertEqual(testcase['script_hash'], str(account.script_hash))
            self.assertIsNotNone(account.public_key)

    def test_new_account_from_encrypted_key(self):
        for testcase in account_list:
            account = Account.from_encrypted_key(testcase['encrypted_key'], testcase['password'])
            self.assertEqual(testcase['address'], account.address)
            self.assertEqual(testcase['encrypted_key'].encode('utf-8'), account.encrypted_key)
            self.assertEqual(testcase['script_hash'], str(account.script_hash))
            self.assertIsNotNone(account.public_key)

    def test_new_watch_only_account(self):
        from neo3.core.types import UInt160
        for testcase in account_list:
            account = Account.watch_only(UInt160.from_string(testcase['script_hash']))
            self.assertEqual(testcase['address'], account.address)
            self.assertIsNone(account.encrypted_key)
            self.assertEqual(testcase['script_hash'], str(account.script_hash))
            self.assertIsNone(account.public_key)

    def test_new_watch_only_account_from_address(self):
        for testcase in account_list:
            account = Account.watch_only_from_address(testcase['address'])
            self.assertEqual(testcase['address'], account.address)
            self.assertIsNone(account.encrypted_key)
            self.assertEqual(testcase['script_hash'], str(account.script_hash))
            self.assertIsNone(account.public_key)

    def test_new_account_from_wif(self):
        for testcase in account_list:
            account = Account.from_wif(testcase['wif_key'], testcase['password'])
            self.assertEqual(testcase['address'], account.address)
            self.assertEqual(testcase['encrypted_key'].encode('utf-8'), account.encrypted_key)
            self.assertEqual(testcase['script_hash'], str(account.script_hash))
            self.assertIsNotNone(account.public_key)

    def test_new_account_wrong_password(self):
        wrong_password: bool = True
        for testcase in account_list:
            try:
                Account.from_encrypted_key(testcase['encrypted_key'], "wrong password")
                wrong_password = False
            except ValueError:
                continue

        self.assertEqual(True, wrong_password)



from typing import Tuple
from neo3 import storage

"""
Internal helpers
"""


class NEOByteCompare:
    def __init__(self, direction="forward"):
        if direction == "forward":
            self.direction = 0
        else:
            self.direction = 1

    def compare(self, x: bytes, y: bytes) -> int:
        if self.direction == 0:
            return self._compare_internal(x, y)
        else:
            return -self._compare_internal(x, y)

    def _compare_internal(self, x: bytes, y: bytes) -> int:
        if x == y:
            return 0
        elif x < y:
            return -1
        else:
            return 1


def NEOSeekSort(comperator_func,
                pair_y: Tuple[storage.StorageKey, storage.StorageItem],
                pair_x: Tuple[storage.StorageKey, storage.StorageItem]) -> int:
    return comperator_func(pair_x[0].key, pair_y[0].key)


def create_find_prefix(id: int, prefix: bytes) -> bytes:
    return id.to_bytes(4, 'little', signed=True) + prefix



from .flightinfo import FlightInfo
from typing import Dict, Optional


class RequestInfo:
    """An internal class for tracking an outstanding header or block request over all connected nodes.

    It is used as part of ``SyncManager`` for syncing the chain over the P2P network.
    For each header or block being requested from the network, 1 RequestInfo object is created. Each RequestInfo stores
    the data to track outstanding flights (=requests from connected nodes).
    """
    def __init__(self, height: int):
        """
        Args:
            height: the header or block height being tracked.
        """
        #: The header or block height being tracked.
        self.height: int = height

        #: A _dictionary holding node :attr:`~neo3.network.node.NeoNode.id` keys, with UTC timestamp values.
        self.failed_nodes: Dict[int, int] = dict()

        #: The total count of flights for this request that failed to meet the time requirements.
        self.failed_total: int = 0

        #: A _dictionary holding node_id keys with :class:`FlightInfo <neo3.network.convenience.flightinfo.FlightInfo>`
        # object values.
        self.flights: Dict[int, FlightInfo] = dict()

        #: The :attr:`~neo3.network.node.NeoNode.id` of the node last used for a flight.
        self.last_used_node: int = -1

    def add_new_flight(self, flight_info: FlightInfo) -> None:
        """
        Store a new flight to the tracking list.

        Args:
            flight_info: the flight to add.
        """
        self.flights[flight_info.node_id] = flight_info
        self.last_used_node = flight_info.node_id

    def most_recent_flight(self) -> Optional[FlightInfo]:
        """
        Get the last :class:`FlightInfo <neo3.network.convenience.flightinfo.FlightInfo>` object created for this
        request.
        """
        try:
            return self.flights[self.last_used_node]
        except KeyError:
            return None

    def mark_failed_node(self, node_id: int) -> None:
        """
        Tag a node for failure.

        SyncManager tags nodes that do not return the requested data before a specified timeout.

        Args:
            node_id: the :attr:`~neo3.network.node.NeoNode.id` of the node the data is requested from.
        """
        self.failed_nodes[node_id] = self.failed_nodes.get(node_id, 0) + 1
        self.failed_total += 1



from unittest import TestCase
from neo3.vm import BigInteger


class BigIntegerTestCase(TestCase):
    def test_big_integer_add(self):
        b1 = BigInteger(10)
        b2 = BigInteger(20)

        b3 = b1 + b2

        self.assertIsInstance(b3, BigInteger)
        self.assertEqual(30, b3)

    def test_big_integer_sub(self):
        b1 = BigInteger(5505505505505505050505)
        b2 = BigInteger(5505505505505505000000)

        b3 = b1 - b2

        self.assertIsInstance(b3, BigInteger)
        self.assertEqual(50505, b3)

    def test_big_integer_mul(self):
        b1 = BigInteger(55055055055055)
        b2 = BigInteger(55055055055)

        b3 = b1 * b2

        self.assertIsInstance(b3, BigInteger)
        self.assertEqual(3031059087112109081053025, b3)

    def test_big_integer_div(self):
        b1 = BigInteger(55055055055055)
        b2 = BigInteger(55055055)

        b3 = b1 / b2
        self.assertIsInstance(b3, BigInteger)
        self.assertEqual(1000000, b3)

    def test_big_integer_div2(self):
        b1 = BigInteger(41483775933600000000)
        b2 = BigInteger(414937759336)

        b3 = b1 / b2
        b4 = b1 // b2
        self.assertIsInstance(b3, BigInteger)
        self.assertEqual(99975899, b3)
        self.assertEqual(b4, b3)

    def test_big_integer_div_rounding(self):
        b1 = BigInteger(1)
        b2 = BigInteger(2)
        self.assertEqual(0, b1 / b2)  # 0.5 -> 0

        b1 = BigInteger(2)
        b2 = BigInteger(3)
        self.assertEqual(0, b1 / b2)  # 0.66 -> 0

        b1 = BigInteger(5)
        b2 = BigInteger(4)
        self.assertEqual(1, b1 / b2)  # 1.25 -> 1

        b1 = BigInteger(5)
        b2 = BigInteger(3)
        self.assertEqual(1, b1 / b2)  # 1.66 -> 1

        b1 = BigInteger(-1)
        b2 = BigInteger(3)
        self.assertEqual(0, b1 / b2)  # -0.33 -> 0

        b1 = BigInteger(-5)
        b2 = BigInteger(3)
        self.assertEqual(-1, b1 / b2)  # -1.66 -> -1

        b1 = BigInteger(1)
        b2 = BigInteger(-2)
        self.assertEqual(0, b1/b2)

    def test_big_integer_div_old_block1473972(self):
        b1 = BigInteger(-11001000000)
        b2 = BigInteger(86400)
        result = b1 / b2
        self.assertEqual(-127326, result)

    def test_big_integer_to_bytearray(self):
        b1 = BigInteger(8972340892734890723)
        ba = b1.to_array()

        integer = BigInteger(ba)
        self.assertEqual(8972340892734890723, integer)

        b2 = BigInteger(-100)
        b2ba = b2.to_array()
        integer2 = BigInteger(b2ba)
        self.assertEqual(-100, integer2)

        b3 = BigInteger(128)
        b3ba = b3.to_array()
        self.assertEqual(b'\x80\x00', b3ba)

        b4 = BigInteger(0)
        b4ba = b4.to_array()
        self.assertEqual(b'\x00', b4ba)

        b5 = BigInteger(-146)
        b5ba = b5.to_array()
        self.assertEqual(b'\x6e\xff', b5ba)

        b6 = BigInteger(-48335248028225339427907476932896373492484053930)
        b6ba = b6.to_array()
        self.assertEqual(20, len(b6ba))

        b7 = BigInteger(-399990000)
        b7ba = b7.to_array()
        self.assertEqual(b'\x10\xa3\x28\xe8', b7ba)

        b8 = BigInteger(-65023)
        b8ba = b8.to_array()
        self.assertEqual(b'\x01\x02\xff', b8ba)

    def test_big_integer_frombytes(self):
        b1 = BigInteger(8972340892734890723)
        ba = b1.to_array()

        b2 = BigInteger(ba)
        self.assertEqual(b1, b2)
        self.assertTrue(b1 == b2)

    def test_big_integer_sign(self):
        b1 = BigInteger(3)
        b2 = BigInteger(0)
        b3 = BigInteger(-4)
        self.assertEqual(1, b1.sign)
        self.assertEqual(0, b2.sign)
        self.assertEqual(-1, b3.sign)

    def test_big_integer_modulo(self):
        b1 = BigInteger(860593)
        b2 = BigInteger(-201)
        self.assertEqual(112, b1 % b2)

        b1 = BigInteger(20195283520469175757)
        b2 = BigInteger(1048576)
        self.assertEqual(888269, b1 % b2)

        b1 = BigInteger(-18224909727634776050312394179610579601844989529623334093909233530432892596607)
        b2 = BigInteger(14954691977398614017)
        self.assertEqual(-3100049211437790421, b1 % b2)

        b3 = BigInteger(b'+K\x05\xbe\xaai\xfa\xd4')
        self.assertEqual(b3, b1 % b2)

    def test_dunder_methods(self):
        b1 = BigInteger(1)
        b2 = BigInteger(2)
        b3 = BigInteger(3)

        self.assertEqual(1, abs(b1))
        self.assertEqual(0, b1 % 1)
        self.assertEqual(-1, -b1)
        self.assertEqual("1", str(b1))
        self.assertEqual(1, b3 // b2)

        right_shift = b3 >> b1
        self.assertEqual(1, right_shift)
        self.assertIsInstance(right_shift, BigInteger)

        left_shift = b1 << b3
        self.assertEqual(8, left_shift)
        self.assertIsInstance(left_shift, BigInteger)

    def test_negative_shifting(self):
        # C#'s BigInteger changes a left shift with a negative shift index,
        # to a right shift with a positive index.

        b1 = BigInteger(8)
        b2 = BigInteger(-3)
        # shift against BigInteger
        self.assertEqual(1, b1 << b2)
        # shift against integer
        self.assertEqual(1, b1 << -3)

        # the same as above but for right shift
        self.assertEqual(64, b1 >> b2)
        self.assertEqual(64, b1 >> -3)

    def test_specials(self):
        self.assertEqual(0, BigInteger.zero())
        self.assertEqual(1, BigInteger.one())
        b = BigInteger.zero()
